/* creates, parses, and prints allnet addresses. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>

#include "types.h"

/* stores an address for sending on the net.
 * if successful, returns the number of bytes used.
 * If bsize is insufficient, returns - the size needed,
 * e.g. -8 for an IPv4 address.
 */
int store_addr (struct allnet_addr * addr, char * buffer, int bsize)
{
  int offset = 2;
  if (bsize >= offset)
    bs2 (buffer, addr->discrim);

  int i, bytes, num_coords, valid_coords;

  switch (addr->discrim) {
  case ALLNET_NO_ADDR_ID:
  case ALLNET_LBCAST_ADDR_ID:
    break;
  case ALLNET_BCAST_ADDR_ID:
    if (bsize > offset)
      bs1 (buffer + offset, b (addr->bc_addr.hops));
    offset++;
    break;
  case ALLNET_KEY_ADDR_ID:
  case ALLNET_UNICAST_ADDR_ID:
    if (bsize > offset)
      bs1 (buffer + offset, b (addr->pk_addr.hops));
    offset++;
    if (bsize > offset)
      bs1 (buffer + offset, b (addr->pk_addr.num_bits));
    offset++;
    bytes = bytes_for_bits (b (addr->pk_addr.num_bits));
/*
    printf ("saving %d bytes, first one is %02x\n", bytes,
            b (addr->pk_addr.pubkey));
*/
    for (i = 0; i < bytes; i++) {
      if (bsize > offset)
        bs1 (buffer + offset, b1 (addr->pk_addr.pubkey + i));
      offset++;
    }
    break;
  case ALLNET_IP4_ADDR_ID:
    if (bsize > offset + 1)
      memcpy (buffer + offset, &(addr->i4_addr.port), 2);
    offset += 2;
    if (bsize > offset + 3)
      memcpy (buffer + offset, &(addr->i4_addr.ip4), 4);
    offset += 4;
    break;
  case ALLNET_IP6_ADDR_ID:
    if (bsize > offset + 1)
      memcpy (buffer + offset     , &(addr->i6_addr.port), 2);
    offset += 2;
    if (bsize >= offset + 3)
      memcpy (buffer + offset + 18, &(addr->i6_addr.scope), 4);
    offset += 4;
    if (bsize > offset + 15)
      memcpy (buffer + offset + 2 , &(addr->i6_addr.ip6), 16);
    offset += 16;
    break;
  case ALLNET_GEO_ADDR_ID:
    num_coords = b (addr->geo_addr.num_coords);
    valid_coords = b (addr->geo_addr.valid_coords);
    if (num_coords < 2 || num_coords > 5) {
      printf ("illegal geographic address with %d coordinates\n", num_coords);
      return 0;
    }
    if (((valid_coords & ALLNET_GEO_LAT_VALID) == 0) ||
        ((valid_coords & ALLNET_GEO_LON_VALID) == 0)) {
      printf ("geographic address lacking lat or lon (0x%02x)\n", valid_coords);
      return 0;
    }
    if (bsize > offset)
      bs1 (buffer + offset, num_coords);
    offset++;
    if (bsize > offset)
      bs1 (buffer + offset, b (addr->geo_addr.valid_coords));
    offset++;

    if (bsize >= offset + 3)
      bs4 (buffer + offset, addr->geo_addr.center_lat);
    offset += 4;
    if (bsize >= offset + 3)
      bs4 (buffer + offset, addr->geo_addr.center_lon);
    offset += 4;

    if (num_coords > 2) {
      int radius = 0;
      if ((valid_coords & ALLNET_GEO_RADIUS_VALID) != 0)
        radius = addr->geo_addr.radius;
      if (bsize >= offset + 3)
        bs4 (buffer + offset, radius);
      offset += 4;
    }
    if (num_coords > 3) {
      int height = 0;
      if ((valid_coords & ALLNET_GEO_HEIGHT_VALID) != 0)
        height = addr->geo_addr.height;
      if (bsize >= offset + 3)
        bs4 (buffer + offset, height);
      offset += 4;
    }
    if (num_coords > 4) {
      int height_range = 0;
      if ((valid_coords & ALLNET_GEO_HEIGHT_RANGE_VALID) != 0)
        height_range = addr->geo_addr.height_range;
      if (bsize >= offset + 3)
        bs4 (buffer + offset, height_range);
      offset += 4;
    }
    break;
  default:
    printf ("error: unknown discriminator %d in store_addr\n", addr->discrim);
    exit (1);
  }
  if (offset <= bsize)
    return offset;
  else
    return -offset;
}

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, the address is stored in *result;
 */
int parse_addr (char * buffer, int bsize, struct allnet_addr * result)
{
  if (bsize < 2) return -1;
  result->discrim = b2 (buffer);
  int expected_bytes = 0;
  int geo_coords = 0;
  int geo_valid = 0;
  switch (result->discrim) {
    case ALLNET_NO_ADDR_ID:
    case ALLNET_LBCAST_ADDR_ID:
      return 2;    /* done */
    case ALLNET_BCAST_ADDR_ID:
      if (bsize < 3) return -1;
      result->bc_addr.hops = b1 (buffer + 2);
      return 3;
    case ALLNET_KEY_ADDR_ID:
    case ALLNET_UNICAST_ADDR_ID:
      if (bsize < 5) return -1;
      result->pk_addr.hops = b1 (buffer + 2);
      result->pk_addr.num_bits = b1 (buffer + 3);
      if (result->pk_addr.num_bits == 0) {
        /* if sending zero bits, use a bcast_addr instead */
        printf ("received key/unicast address with 0 bits, discarding\n");
        return -1;
      }
      expected_bytes = bytes_for_bits (b (result->pk_addr.num_bits));
      if (expected_bytes > 32) return -1;
      if (bsize < 4 + expected_bytes) return -1;
      memcpy (result->pk_addr.pubkey, buffer + 4, expected_bytes);
      return 4 + expected_bytes;
    case ALLNET_IP4_ADDR_ID:
      if (bsize < sizeof (struct allnet_ipv4_addr)) return -1;
      memcpy (&(result->i4_addr.port), buffer + 2, 2);
      memcpy (&(result->i4_addr.ip4), buffer + 4, 4);
      return sizeof (struct allnet_ipv4_addr);
    case ALLNET_IP6_ADDR_ID:
      if (bsize < sizeof (struct allnet_ipv6_addr)) return -1;
      memcpy (&(result->i4_addr.port), buffer + 2, 2);
      memcpy (&(result->i6_addr.scope), buffer + 4, 4);
      memcpy (&(result->i6_addr.ip6), buffer + 8, 16);
      return sizeof (struct allnet_ipv6_addr);
    case ALLNET_GEO_ADDR_ID:
      if (bsize < 3) return -1;
      geo_coords = b1 (buffer + 2);
      if ((geo_coords < 2) || (geo_coords > 5)) {
        printf ("received invalid geographic address with %d coords\n",
                geo_coords);
        return -1;
      }
      expected_bytes = 4 + geo_coords * 4;
      if (bsize < expected_bytes) return -1;
      geo_valid = b1 (buffer + 3);
      result->geo_addr.num_coords = geo_coords;
      result->geo_addr.valid_coords = geo_valid;
      result->geo_addr.center_lat = 0;
      if ((geo_valid & ALLNET_GEO_LAT_VALID) != 0)
        result->geo_addr.center_lat = b4 (buffer + 4);
      result->geo_addr.center_lon = 0;
      if ((geo_valid & ALLNET_GEO_LON_VALID) != 0)
        result->geo_addr.center_lon = b4 (buffer + 8);
      result->geo_addr.radius = 0;
      if ((geo_coords > 2) && ((geo_valid & ALLNET_GEO_RADIUS_VALID) != 0))
        result->geo_addr.radius = b4 (buffer + 12);
      result->geo_addr.height = 0;
      if ((geo_coords > 3) && ((geo_valid & ALLNET_GEO_HEIGHT_VALID) != 0))
        result->geo_addr.height = b4 (buffer + 16);
      result->geo_addr.height_range = 0;
      if ((geo_coords > 4) &&
          ((geo_valid & ALLNET_GEO_HEIGHT_RANGE_VALID) != 0))
        result->geo_addr.height_range = b4 (buffer + 20);
      return expected_bytes;
    default:
      printf ("unknown address type %d/%x\n", result->discrim, result->discrim);
      return -1;
  }
}

/* if the result of calling snprintf is < ssize, return that plus 1, otherwise
 * its negative
 */
#define ret_result(r, s) { if ((r) < 0) \
                             printf ("addr_to_string error: %s\n", (s)); \
                           if ((r) < 0) return 0; \
                           return ((r) + 1) * (((r) >= ssize) ? (-1) : 1); }

/* makes the address printable, adding a '\0' at the end, returning the
 * number of bytes written (including the '\0').
 * If bsize is insufficient, returns - the size needed,
 * e.g. -22 for an IPv4 address such as 123.123.123.123/12345.
 * zero is returned in case of other errors.
 * in case of any error, the string may be garbled, do not use.
 */
int addr_to_string (struct allnet_addr * addr, char * string, int ssize)
{
  int result;
  if (addr == NULL) {
    result = snprintf (string, ssize, "(null addr)\n");
    ret_result (result, "null address");
  }
/*
  printf ("address type %d (%x) for %p\n", addr->discrim, addr->discrim, addr);
*/
  int i, number, bytes, needed, hops;
  char * addr_type_name = "pub key";
  /* for printing internet addresses.  Size is too big, which is OK */
  static char ipaddr [INET6_ADDRSTRLEN + INET_ADDRSTRLEN + 1];
  double lat, lon, radius, height, hr;
  switch (addr->discrim) {
  case ALLNET_NO_ADDR_ID:
    result = snprintf (string, ssize, "no address");
    ret_result (result, "ALLNET_NO_ADDR_ID");
  case ALLNET_LBCAST_ADDR_ID:
    result = snprintf (string, ssize, "local broadcast address");
    ret_result (result, "local broadcast");
  case ALLNET_BCAST_ADDR_ID:
    result = snprintf (string, ssize, "broadcast address, %d hops max",
                       b (addr->bc_addr.hops));
    ret_result (result, "bcast");
  case ALLNET_KEY_ADDR_ID:
  case ALLNET_UNICAST_ADDR_ID:
    hops = b (addr->pk_addr.hops);
    number = b (addr->pk_addr.num_bits);
    if (addr->discrim == ALLNET_UNICAST_ADDR_ID) addr_type_name = "unicast";
    result = snprintf (string, ssize, "%s, %d hops, %d bits: ",
                       addr_type_name, hops, number);
    if (result < 0) ret_result (result, "insufficient space for key/unicast");
    if (result + 1 < ssize) { string += result; ssize -= result; }
    needed = result;
    bytes = bytes_for_bits (number);
    for (i = 0; i < bytes; i++) {
      char * follows = ", ";
      if (i + 1 == bytes) follows = "";
      result = snprintf (string, ssize, "0x%02x%s",
                         b1 (addr->pk_addr.pubkey + i), follows);
      if (result < 0) ret_result (result, "insufficient space (2) for k/u"); 
      if (needed + result + 1 < ssize) { string += result; ssize -= result; }
      needed += result;
    }
    /* result = snprintf (string, ssize, "\n"); */
    if (result < 0) ret_result (result, "insufficient space (3) for k/u"); 
    needed += result;
    ret_result (needed, "key/unicast");
  case ALLNET_IP4_ADDR_ID:
    if (inet_ntop (AF_INET, &(addr->i4_addr.ip4), ipaddr, sizeof (ipaddr))
        == NULL) {
      perror ("inet_ntop/4");
      printf ("unprintable ipv4, port %d/%x\n",
              ntohs (addr->i4_addr.port), ntohs (addr->i4_addr.port));
      ret_result (result, "unprintable ip4");
    }
    result = snprintf (string, ssize, "ipv4 address %s, port %d/%x",
                       ipaddr, ntohs (addr->i4_addr.port),
                       ntohs (addr->i4_addr.port));
    ret_result (result, "ip4");
  case ALLNET_IP6_ADDR_ID:
    if (inet_ntop (AF_INET6, &(addr->i6_addr.ip6), ipaddr, sizeof (ipaddr))
        == NULL) {
      perror ("inet_ntop/6");
      printf ("unprintable ipv6, port %d/%x\n", ntohs (addr->i6_addr.port),
              ntohs (addr->i6_addr.port));
      ret_result (result, "unprintable ip6");
    }
    result = snprintf (string, ssize, "ipv6 address %s, port %d/%x, scope %d",
                       ipaddr, ntohs (addr->i6_addr.port),
                       ntohs (addr->i6_addr.port), addr->i6_addr.scope);
    ret_result (result, "ip6");
  case ALLNET_GEO_ADDR_ID:
    lat = ((double) addr->geo_addr.center_lat) * 180.0;
    lon = ((double) addr->geo_addr.center_lon) * 180.0;
    result = snprintf (string, ssize, "(%d/%02x) lat %f, lon %f",
                       b (addr->geo_addr.num_coords),
                       b (addr->geo_addr.valid_coords), lat, lon);
    if (result < 0) ret_result (result, "geo address, not enough space");
    if (result + 1 < ssize) { string += result; ssize -= result; }
    needed = result;
    if ((addr->geo_addr.num_coords > 2) &&
        ((addr->geo_addr.valid_coords & ALLNET_GEO_RADIUS_VALID) != 0)) {
      result = snprintf (string, ssize, ", radius %f",
                         ((double) addr->geo_addr.radius) * 180.0);
      if (result < 0) ret_result (result, "geo address/2, not enough space");
      if (needed + result + 1 < ssize) { string += result; ssize -= result; }
      needed += result;
    }
    if ((addr->geo_addr.num_coords > 3) &&
        ((addr->geo_addr.valid_coords & ALLNET_GEO_HEIGHT_VALID) != 0)) {
      result = snprintf (string, ssize, ", height %.2f",
                         ((double) addr->geo_addr.height) / 100.0);
      if (result < 0) ret_result (result, "geo address/3, not enough space");
      if (needed + result + 1 < ssize) { string += result; ssize -= result; }
      needed += result;
    }
    if ((addr->geo_addr.num_coords > 4) &&
        ((addr->geo_addr.valid_coords & ALLNET_GEO_HEIGHT_RANGE_VALID) != 0)) {
      result = snprintf (string, ssize, " +- %.2f",
                         ((double) addr->geo_addr.height_range) / 100.0);
      if (result < 0) ret_result (result, "geo address/4, not enough space");
      if (needed + result + 1 < ssize) { string += result; ssize -= result; }
      needed += result;
    }
    result = snprintf (string, ssize, "\n");
    if (result < 0) ret_result (result, "geo address/5, not enough space");
    needed += result;
    ret_result (needed, "geo");
  default:
    printf ("unknown address type %d in addr_to_string\n", addr->discrim);
    char * p = NULL;
    *p = 3;
    return 0;
  }
  printf ("error: got to the end of addr_to_string\n");
  return 0;
}

/* prints the address, optionally with a newline at the end (if eof > 0) */
/* malloc's the buffer so it is suitable for multithreaded code */
void print_addr (struct allnet_addr * addr, int eol)
{
  int bufsize = 10000;
  char * buffer = malloc (bufsize);
  if (buffer != NULL) {
    int printed = addr_to_string (addr, buffer, bufsize);
    if (printed > 0) {
      if (eol > 0)
        printf ("%s\n", buffer);
      else
        printf ("%s", buffer);
    } else {
      printf ("error printing address of type %d, result %d\n", addr->discrim,
              printed);
    }
    free (buffer);
  } else {
    printf ("memory error, unable to print address\n");
  }
}

/* returns 1 if the addresses are the same, 0 if they differ */
int same_addr (struct allnet_addr * a1, struct allnet_addr * a2)
{
  if (a1->discrim != a2->discrim)
    return 0;
  int i, n, bits, bytes;
  switch (a1->discrim) {
  case ALLNET_NO_ADDR_ID:
  case ALLNET_LBCAST_ADDR_ID:
    return 1;                    /* same address */
  case ALLNET_BCAST_ADDR_ID:
    /* two broadcast addresses are always the same, even if the number of
     * hops is different. */
    return 1;                    /* same address */
  case ALLNET_KEY_ADDR_ID:
  case ALLNET_UNICAST_ADDR_ID:
    /* the number of hops does not affect equality */
    if (b (a1->pk_addr.num_bits) != b (a2->pk_addr.num_bits)) return 0;
    n = bytes_for_bits (b (a1->pk_addr.num_bits));
    /* for the first n - 1 bytes, the entire byte should match */
    for (i = 0; i + 1 < n; i++)
      if (b1 (a1->pk_addr.pubkey + i) != b1 (a2->pk_addr.pubkey + i)) return 0;
    /* compare the first few bits of the last byte */ 
    bits = (b (a1->pk_addr.num_bits)) % 8;
    if (bits == 0) bits = 8;
    printf ("bytes 0..%d are the same, last byte (%d bits) is %x =? %x\n",
            n - 2, bits,
            b1 (a1->pk_addr.pubkey + n - 1), b1 (a2->pk_addr.pubkey + n - 1));
    /* 8 >= bits > 0 */
    if ((b1 (a1->pk_addr.pubkey + n - 1) >> (8 - bits)) !=
        (b1 (a2->pk_addr.pubkey + n - 1) >> (8 - bits)))
      return 0;
    printf ("last %d bits are the same\n", bits);
    return 1;                    /* same address */
  case ALLNET_IP4_ADDR_ID:
    printf ("comparing addresses %d to %d\n",
            a1->i4_addr.ip4.s_addr, a2->i4_addr.ip4.s_addr);
    if (a1->i4_addr.ip4.s_addr != a2->i4_addr.ip4.s_addr) return 0;
    if (a1->i4_addr.port != a2->i4_addr.port) return 0;  /* different */
    printf ("addresses %d and %d are the same\n",
            a1->i4_addr.ip4.s_addr, a2->i4_addr.ip4.s_addr);
    return 1;                    /* same address */
  case ALLNET_IP6_ADDR_ID:
    if ((a1->i6_addr.ip6.s6_addr32 [0] != a2->i6_addr.ip6.s6_addr32 [0]) ||
        (a1->i6_addr.ip6.s6_addr32 [1] != a2->i6_addr.ip6.s6_addr32 [1]) ||
        (a1->i6_addr.ip6.s6_addr32 [2] != a2->i6_addr.ip6.s6_addr32 [2]) ||
        (a1->i6_addr.ip6.s6_addr32 [3] != a2->i6_addr.ip6.s6_addr32 [3]))
      return 0;
    if (a1->i6_addr.port != a2->i6_addr.port) return 0;  /* different */
    /* not sure if we should compare scopes, but why not */
    if (a1->i6_addr.scope != a2->i6_addr.scope) return 0;  /* different */
    return 1;                    /* same address */
  case ALLNET_GEO_ADDR_ID:
    if ((a1->geo_addr.num_coords != a2->geo_addr.num_coords) ||
        (a1->geo_addr.valid_coords != a2->geo_addr.valid_coords))
      return 0;
    if (a1->geo_addr.center_lat != a2->geo_addr.center_lat) return 0;
    if (a1->geo_addr.center_lon != a2->geo_addr.center_lon) return 0;
    if ((a1->geo_addr.num_coords > 2) &&
        ((a1->geo_addr.valid_coords & ALLNET_GEO_RADIUS_VALID) != 0))
      if (a1->geo_addr.radius != a2->geo_addr.radius) return 0;
    if ((a1->geo_addr.num_coords > 3) &&
        ((a1->geo_addr.valid_coords & ALLNET_GEO_HEIGHT_VALID) != 0))
      if (a1->geo_addr.height != a2->geo_addr.height) return 0;
    if ((a1->geo_addr.num_coords > 4) &&
        ((a1->geo_addr.valid_coords & ALLNET_GEO_HEIGHT_RANGE_VALID) != 0))
      if (a1->geo_addr.height_range != a2->geo_addr.height_range) return 0;
    return 1;                    /* same address */
  default:
    return 0;                   /* may be same address, but how do I know? */
  }
}

/* returns 1 if the address 1 matches address 2, 0 if it cannot. */
/* this is mostly the same as same_addr, but for keys and unicast
 * it allows a more generic a1 to match a more specific a2.
 * for geometric addresses it returns whether there is any overlap */
int matches_addr (struct allnet_addr * a1, struct allnet_addr * a2)
{
  if (a1->discrim != a2->discrim)
    return 0;
  if (same_addr (a1, a2))
    return 1;
  /* not the same address */
  /* only have to worry about keys/unicast and geo addresses */
  int i, n, bits, bytes;
  switch (a1->discrim) {
  case ALLNET_KEY_ADDR_ID:
  case ALLNET_UNICAST_ADDR_ID:
    /* allow a1 to have fewer bits than a2, but not the reverse */
    if (a1->pk_addr.num_bits > a2->pk_addr.num_bits) return 0;
    /* only compare the number of bits specified by a1 */
    n = bytes_for_bits (a1->pk_addr.num_bits);
    for (i = 0; i + 1 < n; i++)
      if (a1->pk_addr.pubkey [i] != a2->pk_addr.pubkey [i]) return 0;
    /* compare the first few bits of the last byte */ 
    bits = (a1->pk_addr.num_bits & 0xff) % 8;
    if (bits == 0) bits = 8;
/*
    printf ("bytes 0..%d match, last byte (%d bits) is %x =? %x\n",
            n - 2, bits,
            b1 (a1->pk_addr.pubkey + n - 1), b1 (a2->pk_addr.pubkey + n - 1));
 */
    /* 8 >= bits > 0 */
    if ((b1 (a1->pk_addr.pubkey + n - 1) >> (8 - bits)) !=
        (b1 (a2->pk_addr.pubkey + n - 1) >> (8 - bits)))
      return 0;
/*
    printf ("last %d bits match\n", bits);
 */
    return 1;                    /* matching address */
  case ALLNET_GEO_ADDR_ID:
    printf ("matching non-identical geographic addresses not implemented\n");
    return 0;
  default:
    return 0;                   /* may match, but how do I know? */
  }
}
