/* interface to allnet functions */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "types.h"
#include "messages.h"
#include "api.h"
#include "hash.h"

/* allnet_start should be called before using allnet, and allnet_stop
 * should be called after finishing with allnet.
 *
 * allnet_start opens and returns a given port for communication with
 * the allnet daemon.
 *
 * whenever a message is available, the callback function is called.
 * the callback function must complete execution and return before
 * the next message will be delivered.
 *
 * all the parameters passed to the callback function may be reused
 * by allnet on the next call, so should not be saved anywhere.
 *
 * the callback function may call allnet_stop, allnet_send, allnet_ack
 *
 * the caller may provide an arbitrary pointer ("arg") which will
 * be given to the callback function on each call.
 *
 * the return value from allnet_start is a port number >= 0 (success),
 * otherwise -1 (failure).
 */

struct port_info {
  int in_use;
  int socket;
  allnet_receive_fun callback;
  void * callback_arg;
  char * buffer;
  int bsize;
  struct allnet_addr addr;
  struct allnet_route route;
  pthread_t receive_thread;
  int stop_receiving;
};

#define MAX_PORTS	10
struct port_info open_ports [MAX_PORTS];
int initialized = 0;

void * receive_loop (void * arg);

/*
typedef void (* allnet_receive_fun) (void * arg, int port,
                                     char * buffer, int bsize,
                                     struct allnet_addr * from,
                                     struct allnet_route * route, int key_id);
*/
int allnet_start (allnet_receive_fun callback, void * arg)
{
  if (! initialized) {
    int i;
    for (i = 0; i < MAX_PORTS; i++)
      open_ports [i].in_use = 0;
    initialized = 1;
  }

  int index = 0;
  while ((index < MAX_PORTS) && (open_ports [index].in_use))
    index++;
  if (index >= MAX_PORTS) {
    printf ("no more allnet ports available\n");
    return -1;
  }

  struct port_info * p = open_ports + index;

/* do the easy initializations first */
  p->callback = callback;
  p->callback_arg = arg;

  p->addr.discrim = ALLNET_NO_ADDR_ID;

  p->route.num_addrs = 0;
  p->route.addrs = NULL;

  p->stop_receiving = 0;

  p->socket = socket (AF_INET, SOCK_DGRAM, 0);
  if (p->socket < 0) {
    perror ("socket");
    return -1;
  }

  p->bsize = 66000;
  p->buffer = malloc (p->bsize);
  if (p->buffer == NULL) {
    printf ("unable to create receive buffer\n");
    return -1;
  }

  if (pthread_create (&(p->receive_thread), NULL,
                      receive_loop, (void *) index) != 0) {
    free (p->buffer);
    printf ("unable to create receive thread\n");
    return -1;
  }

  /* send an empty message to the allnet daemon to let them know we are here */
  struct sockaddr_in sin;
  sin.sin_family = AF_INET;
  /* set the loopback address */
  sin.sin_addr.s_addr = htonl (0x7F000001);
  sin.sin_port = ALLNET_LOCAL_PORT;
  /* should we even check the return value? */
  if (sendto (p->socket, p->buffer, 0, 0,
              (struct sockaddr *) (&sin), sizeof (sin)) < 0)
    perror ("initial sendto");

  /* commit the new port */
  p->in_use = 1;

  return index + 1;
}

void allnet_stop (int port)
{
  if ((port > 0) && (port <= MAX_PORTS))
    open_ports [port - 1].stop_receiving = 1;
/* the receive thread does all the shutting down before exiting */
}

void check_end_thread (int port)
{
  struct port_info * p = open_ports + (port - 1);
  if (p->stop_receiving) {
    close (p->socket);
    free (p->route.addrs);
    free (p->buffer);
    p->in_use = 0;
    pthread_exit (NULL);
  }
}

void * receive_loop (void * arg)
{
  int index = (int) arg;
  struct port_info * p = open_ports + index;
  int flags = MSG_DONTWAIT | MSG_TRUNC;
  struct sockaddr_storage sa;
  struct sockaddr * sap = (struct sockaddr *) &sa;
  while (1) {
    socklen_t al = sizeof (sa);
#ifdef DEBUG_PACKETS
    printf ("receive_loop on port %d (%d, %p, %d, %x, %p, %d)\n",
            index, p->socket, p->buffer, p->bsize, flags, sap, al);
#endif /* DEBUG_PACKETS */
    int received;
    do {
      received = recvfrom (p->socket, p->buffer, p->bsize, flags, sap, &al);
      if ((received == -1) && ((errno == EAGAIN) || (errno == EWOULDBLOCK))) {
        received = -2;
        struct timespec t;
        t.tv_sec = 0;
        t.tv_nsec = 10 * 1000 * 1000;   /* sleep 10ms or so */
        nanosleep (&t, NULL);
      }
    } while (received == -2);
    if (received == -1) {
      perror ("recvfrom");
      continue;   /* receive again */
    }
#ifdef DEBUG_PACKETS
    printf ("receive_loop on port %d received %d bytes\n", index, received);
#endif /* DEBUG_PACKETS */
    check_end_thread (index);
    if (received == 1) { /* keepalive packet, send it right back */
      if (sendto (p->socket, p->buffer, 1, 0, sap, al) < 0)
        perror ("sendto/keepalive");
      continue;  /* restart loop */
    }
    if (received > p->bsize) {
      printf ("received message had %d bytes, last %d bytes lost\n", received,
              received - p->bsize);
      received = p->bsize;
    }
    struct allnet_max_header header_storage;
    struct allnet_header * hp = (struct allnet_header *) (&header_storage);
    char * owner_info;
    int offset = parse_header (p->buffer, received, hp, &owner_info);
    if (offset < 0) {
      printf ("unable to parse received buffer\n");
      continue;   /* receive again */
    }
    if ((b (hp->allnet_addrs_used)) > (b (hp->allnet_num_addrs))) {
      printf ("received header has %d addresses, uses %d\n",
              b (hp->allnet_num_addrs), b (hp->allnet_addrs_used));
      continue;   /* receive again */
    }
    struct allnet_addr from = hp->route [b (hp->allnet_addrs_used)];
#ifdef DEBUG_PACKETS
    printf (" from ");
    print_addr (&from, 1);
#endif /* DEBUG_PACKETS */
    int size = received - offset;
    char * data = p->buffer + offset;
    int call_app = 1;

    /* try to decrypt the packet */
    char * key_hint = NULL;
    int chosen_key_id = -1;
    if (hp->allnet_type == ALLNET_TYPE_DATA) {
      int hint_size = 0;
      if (from.discrim == ALLNET_KEY_ADDR_ID) {
        hint_size = (from.pk_addr.num_bits & 0xff) / 8;
        if (hint_size > sizeof (from.pk_addr.pubkey))
          hint_size = sizeof (from.pk_addr.pubkey);
        if (hint_size > 0)
          key_hint = from.pk_addr.pubkey;
      }
      char * text_buffer = malloc (size);
#ifdef DEBUG_PACKETS
      printf ("decrypting message with key hint %d bits, using %d bytes\n",
              b (from.pk_addr.num_bits), hint_size);
#endif /* DEBUG_PACKETS */
      if (text_buffer != NULL) {
        int decoded = 
          decrypt_verify (data, size, text_buffer, size, key_hint, hint_size,
                          &chosen_key_id);
        if (decoded > size) {
          printf ("decoded message %d larger than received %d\n",
                  decoded, size);
          decoded = size;
        }
        if (decoded == -1) {
          /* too many to print, including my own messages sent back to me
           printf ("unable to decrypt incoming message, discarding\n"); */
          call_app = 0;
        } else {
          memcpy (data, text_buffer, decoded);
          size = decoded;
        }
        free (text_buffer);
      }
    } else if (hp->allnet_type == ALLNET_TYPE_CLEAR) {
      /* deliver the message to the app with key_id == -1 */
    } else {
      printf ("got message of type %d, not delivering to app\n",
              hp->allnet_type);
      call_app = 0;
    }

    if (call_app)
      p->callback (p->callback_arg, index + 1, data, size,
                   hp->route, b (hp->allnet_num_addrs),
                   b (hp->allnet_addrs_used), chosen_key_id);
#ifdef DEBUG_PACKETS
    if (call_app)
      printf ("callback done, port %d, key %d\n", index + 1, chosen_key_id);
#endif /* DEBUG_PACKETS */
    check_end_thread (index);
  }
}

/* sends the message.
 *
 * If the address is NULL, the message will be sent using
 * a broadcast, otherwise using the given address.
 *
 * If the key is 0, the message will be sent in the clear,
 * otherwise sent encrypted with the given key.
 *
 * if osize is zero, owner_id may be NULL
 * otherwise, the owner ID is always put into the message, and if
 * send_owner_id_hash is true, the hash of the owner ID is sent in the header
 */
char * allnet_encrypt (char * message, int mlen,
                       int key_id, char * owner_id, int osize,
                       int * result_size, int nbits, char * key_bits)
{
/* obsolete, I think
  if (key_id == 0) {
    printf ("allnet_encrypt error: key ID should not be 0\n");
    return NULL;
  }
*/
  if ((osize != 0) && (osize != OWNER_ID_SIZE)) {
    printf ("illegal owner_id size %d, must be %d\n", osize, OWNER_ID_SIZE);
    return NULL;
  }
  if (osize == 0)
    owner_id = NULL;

  /* 1024 is normally enough for short messages, but keys may grow later */
  int clen = 4096;
  char * cypher = malloc (clen);

  int save_size = 0;
  if ((nbits > 0) && (key_bits != NULL))
    save_size = (nbits + 7) / 8;
 
  int s = sign_encrypt (message, mlen, cypher, clen, owner_id,
                        key_id, key_bits, save_size);
  if (s == -1) {
    printf ("unable to encrypt message '%s'\n", message);
    return NULL;
  }
  *result_size = s;
  return cypher;
}

char * allnet_ack (char * original_message, int mlen,
                   int * result_size)
{
  if (mlen < OWNER_ID_SIZE) {
    printf ("unable to send ack, message length %d, %d needed\n",
            mlen, OWNER_ID_SIZE);
    return NULL;
  }
  int rsize = CLEAR_ID_SIZE + OWNER_ID_SIZE;
  char * result = malloc (rsize);
  memcpy (result, ACKNOWLEDGEMENT_BYTES, CLEAR_ID_SIZE);
  memcpy (result + CLEAR_ID_SIZE, original_message, OWNER_ID_SIZE);
  *result_size = rsize;
  return result;
}

/* type 0 is an encrypted message, 1 is a clear message */
static void allnet_send_msg (int port, char * message, int mlen, int type,
                             struct allnet_addr * route, int num_addrs,
                             int used,
                             char * owner_id, int osize)
{
  if ((port <= 0) || (port > MAX_PORTS)) {
    printf ("allnet_send, illegal port number %d\n", port);
    return;
  }
  struct port_info * p = open_ports + (port - 1);

  struct allnet_addr to;
  if (route != NULL) {
    if ((num_addrs <= 0) || (used < 0) || (used + 1 > num_addrs)) {
      printf ("illegal call to allnet_send, %d addrs, %d used\n",
              num_addrs, used);
      return;
    }
    to = route [used];
  } else {
    to.discrim = ALLNET_BCAST_ADDR_ID;
    to.bc_addr.hops = 6;
    route = &to;
    num_addrs = 1;
    used = 0;
  }
  char owner_id_hash_storage [HASH_SIZE];
  char * owner_id_hash = NULL;
  int hash_size = 0;
  if ((owner_id != NULL) && (osize > 0)) {
    allnet_hash (owner_id, osize, owner_id_hash_storage);
    hash_size = HASH_SIZE;
    if (hash_size > OWNER_ID_SIZE)
      hash_size = OWNER_ID_SIZE;
    owner_id_hash = owner_id_hash_storage;
  }

  /* get the size first */
  int size = - store_data (message, mlen, type, route, num_addrs, used,
                           owner_id_hash, hash_size, NULL, 0);
  char * buffer = malloc (size);
  if (buffer == NULL) {
    printf ("unable to allocate buffer for message to send, sizes %d/%d\n",
            mlen, size);
    return;
  }
  /*
  printf ("using buffer %p of size %d to send %d bytes of user data\n",
          buffer, size, mlen);
  */

  int store_result = store_data (message, mlen, type, route, num_addrs, used,
                                 owner_id_hash, hash_size, buffer, size);
  if (store_result != size) {
    printf ("error: unknown result %d from store_data, %d expected\n",
            store_result, size);
    exit (1);
  }

  /* send to allnetd */
  struct sockaddr_in sin;
  sin.sin_family = AF_INET;
  /* set the loopback address */
  sin.sin_addr.s_addr = htonl (0x7F000001);
  sin.sin_port = ALLNET_LOCAL_PORT;
  /* printf ("sending %d bytes\n", mlen); */
  if (sendto (p->socket, buffer, size, 0,
              (struct sockaddr *) (&sin), sizeof (sin)) < 0)
    perror ("sendto");
  free (buffer);
}

void allnet_send (int port, char * message, int mlen,
                  struct allnet_addr * route, int num_addrs, int used,
                  char * owner_id, int osize)
{
  allnet_send_msg (port, message, mlen, 0, route, num_addrs, used,
                   owner_id, osize);
}

void allnet_send_clear (int port, char * message, int mlen,
                        struct allnet_addr * route, int num_addrs, int used)
{
  allnet_send_msg (port, message, mlen, 1, route, num_addrs, used, NULL, 0);
}

