/* interface to allnet functions */

#ifndef ALLNET_API_H
#define ALLNET_API_H

#include "types.h"

#define ALLNET_API_DIRECTORY	".allnet"

/* allnet_start should be called before using allnet, and allnet_stop
 * should be called after finishing with allnet.
 *
 * allnet_start opens and returns a given port for communication with
 * the allnet daemon.
 *
 * whenever a message is available, the callback function is called.
 * the callback function must complete execution and return before
 * the next message will be delivered.
 *
 * all the parameters passed to the callback function may be reused
 * by allnet on the next call, so should not be saved anywhere.
 *
 * the callback function may call allnet_stop or allnet_send
 *
 * key_id is set to the key ID for a successfully decrypted message,
 * and to 0 for an unencrypted message.
 *
 * the caller may provide an arbitrary pointer ("arg") which will
 * be given to the callback function on each call.
 *
 * the return value from allnet_start is a port number >= 0 (success),
 * otherwise -1 (failure).
 */
typedef void (* allnet_receive_fun) (void * arg, int port,
                                     char * buffer, int bsize,
                                     struct allnet_addr * route,
                                     int route_num_addr, int route_used,
                                     int key_id);
extern int allnet_start (allnet_receive_fun callback, void * arg);
extern void allnet_stop (int port);

/* functions implemented in allnet_api.c */
/* allocates and creates a message for later use with allnet_send.
 * the message must be free'd when it is no longer needed.
 *
 * If the key is 0, the message will be sent in the clear,
 * otherwise sent encrypted with the given key.
 *
 * if osize is zero, owner_id may be NULL
 * otherwise, the owner ID is always put into the message, and if
 * send_owner_id_hash is true, the hash of the owner ID is sent in the header
 */
extern char * allnet_encrypt (char * message, int mlen,
                              int key_id, char * owner_id, int osize,
                              int * result_size, int nbits, char * key_bits);

/* acknowledge the message by sending the first OWNER_ID_SIZE of
 * the original message back along the reverse route.
 */
extern char * allnet_ack (char * original_message, int mlen,
                          int * result_size);

/* send a complete message, as generated by allnet_encrypt or allnet_ack */
/* the route given is used to build the header -- if it is NULL,
 * the message is broadcast.
 * if owner_id_hash is not NULL, it must refer to osize bytes
 * that will be hashed and used as an owner ID
 */
extern void allnet_send (int port, char * message, int mlen,
                         struct allnet_addr * route, int num_addrs, int used,
                         char * owner_id, int osize);

/* same, but send in the clear */
extern void allnet_send_clear (int port, char * message, int mlen,
                               struct allnet_addr * route, int num_addrs,
                               int used);

/* functions for key management, implemented in key.c */

/* there is a separate set of keys for each contact in the chat program */
/* each set of keys consists of my private and public keys, and the contact's
 * public key, saved in files 'priv', 'pub', and 'contact'.
 * The contact's name is saved in file 'name'
 * All other files in each contact's directory are content files, to be
 * managed by the application.
 */

/* returns the number n of contacts if successful, otherwise -1.
 *
 * if no keys were found, returns 0, but can use
 * keys if they are added with add_key.
 *
 * valid key IDs have values from 1..n */
extern int init_key_files (char * path, char * program_name);

/* returns 1 if it found a name for the given key.  Puts as much of the
 * name as possible into namebuf (and always terminates the string, as
 * long as bsize > 0).  If not found, returns 0, -1 for errors.
 * if a key does not have a name, namebuf is set to "",
 * but name_for_key still returns 1.
 *
 * if path is not null and psize > 0, also fills in the path to a
 * directory where user information can be saved.  If the path does
 * not fit in psize bytes (including the terminating null character),
 * sets path to the empty string.  Otherwise, path is always filled
 * in if the return value is 1.
 */
extern int name_for_key (int key_id, char * namebuf, int bsize,
                         char * path, int psize);

/* given any part of a name, returns all the matching key IDs, up to
 * num_keys.  Also returns the total number of matching keys, which
 * may be greater than num_keys.  Returns -1 in case of errors.
 */
extern int key_for_name (char * name, int * key_id, int num_keys);

/* given a key ID, returns the first few bytes of the public key
 * returns the number of bytes returned, never more than num_bytes,
 * or -1 for errors */
extern int key_bytes (int key_id, char * key_bytes, int num_bytes);

/* adds the given key and stores it in the file.
 * returns the key_id of the new key, or -1 in case of failure.
 */
extern int add_key (char * name, char * key, int key_len);

/* creates the new key, and saves it.  The key is not really available
 * until add_key is called.
 * returns 1 for success, 0 for failure.
 */
extern int new_keypair ();

/* fills in cypher with the signed and encrypted text.  key_id is the
 * ID identifying the recipient.  Also fills in the first ksize
 * bytes of the key, padding with zeros if ksize > key length.
 * returns the number of bytes in the cyphertext if successful -- if
 * the value returned > csize, the result does not fit and is not
 * stored in cypher.
 * owner_id is either NULL, or a pointer to OWNER_ID_SIZE bytes.
 * if owner_id is not NULL, owner_id_hash must also be a pointer
 * to OWNER_ID_SIZE bytes.
 * returns -1 in case of failure.
 */
extern int sign_encrypt (char * text, int tsize, char * cypher, int csize,
                         char * owner_id, int key_id, char * key, int ksize);

/* if successfully decrypts and verifies, fills in text (up to tsize) and
 * returns the total number of characters in text (which may exceed tsize).
 * key_hint, if not NULL, is used to restrict the number of personal keys
 * which need to be tried.
 *
 * if it finds a matching key, sets *key_id (if not NULL)
 *
 * if unable to find a key to decrypt and verify, returns -1
 */
extern int decrypt_verify (char * cypher, int csize, char * text, int tsize,
                           char * key_hint, int ksize, int * key_id);

/* result is malloc'd, should be free'd */
extern char * random_string (int chars);

/* creates a message to be sent as part of a key exchange.
 * string must be shared among the two sides and must be hard for
 * an attacker to guess -- recommend a random string at least 10
 * letters long.
 *
 * storage for the message is allocated dynamically, and the caller
 * must free() the result of this call.  The length of the resulting
 * message is stored in len (the message is not null-terminated).
 *
 * The way this works is:
 * each side chooses a public key K and a nonce N.
 * Both sides must share a string S which is hard to guess.
 *
 * Then, each side computes a hash H = hash (S, K, N), and sends to
 * the peer (H, K, N).  The receiver receiving (H', K', N') can
 * recompute H' = hash (S, K', N').
 *
 * If H equals H', and if S is hard to guess, the receiver can trust
 * that the sender knows S, and that K' equals K and is the correct
 * public key for that sender.
 *
 * The nonce is useful in case the user reuses S, the nonce makes
 * that fact less obvious.  Still, reusing S exposes the users to
 * replay attacks, and is a bad idea.  A randomly generated S is
 * much better.  Use random_string() to suggest a string to the user.
 */
extern char * message_for_key_exchange (char * string, int * len);

/* verifies whether the received message matches the given string.
 * returns -1 in case of no match, and the new key ID in case of success.
 * If the new key matches an existing key, returns the key ID of the
 * existing key.  The key is saved with "key_name" as the name for the key.
 */
extern int verify_key_exchange (char * msg, int len, char * string,
                                char * key_name);

/* use may_call_random != 0 if this function may call random(), i.e. is
 * not used to set a seed for the built-in random number generator.
 * use alpha != 0 if the result should be letters a-z (minus i and o)
 */
extern void random_fill (char * buffer, int bytes,
                         int may_call_random, int alpha);

#endif /* ALLNET_API_H */
