
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <dirent.h>
#include <openssl/rsa.h>
#include <openssl/pem.h>
#include <openssl/objects.h>
#include <openssl/err.h>
#include <openssl/ripemd.h>

#include "api.h"
#include "hash.h"
#include "messages.h"

/* apparently not always in string.h */
char *strcasestr(const char *haystack, const char *needle);

/* return -1 in case of error, or a file descriptor otherwise.
 * if file is not NULL, also fdopens the file descriptor and assigns
 * the FILE* pointer to *file
 * if size is not null, fills in the size of the file
 */
static int open_create_file (char * fname, off_t * size, FILE * * file)
{
  int mode = S_IRWXU;
  int fd = open (fname, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
  if (fd < 0) {
    perror ("open");
    return -1;
  }
  if (size != NULL) {
    struct stat info;
    if (fstat (fd, &info) < 0) {
      perror ("fstat");
      close (fd);
      return -1;
    }
    *size = info.st_size;
  }
  if (file != NULL) {
    *file = fdopen (fd, "a+");
    if (*file == NULL) {
      perror ("fdopen");
      close (fd);
      return -1;
    }
  }
  return fd;
}

static char * malloc_strcpy (char * string)
{
  int size = strlen (string) + 1;
  char * result = malloc (size);
  if (result == NULL) {
    printf ("unable to malloc %d bytes to copy string '%s'\n", size, string);
    return NULL;
  }
  strcpy (result, string);
  return result;
}

/* append to the end of a file and print any errors.
 * Returns 1 in case of success, 0 error */
static int my_append (int fd, char * buffer, int bytes, char * description)
{
  if (lseek (fd, 0, SEEK_END) == ((off_t) -1)) {
    perror ("lseek in my_append");
    if (description != NULL)
      printf ("%s seek error", description);
    return 0;
  }
  int result = write (fd, buffer, bytes);
  if (result != bytes) {
    if (result < 0)
      perror ("write");
    if (description != NULL)
      printf ("%s error: ", description);
    printf ("wrote %d bytes, actually wrote %d bytes\n", bytes, result);
    return 0;
  }
  return 1;
}

/* return -1 in case of error, or the number of bytes read otherwise.
 * linep (if not NULL) is set to point to a newly allocated buffer
 * containing the line
 */
static int my_readline (FILE * file, char ** linep)
{
  long start = ftell (file);
  int count = 0;
  while (1) {
    int c = fgetc (file);
    if (c == EOF) { /* no newline found */
      if (linep != NULL)
        *linep = NULL;
      return -1;
    }
    if (c == '\n')
      break;
    count++;
  }
  if ((linep == NULL) || (count == 0))
    return count;
  char * line = malloc (count + 1);
  *linep = line;
  if (line == NULL) {
    printf ("unable to allocate line for key\n");
    return -1;
  }
  fseek (file, start, SEEK_SET);  /** go back and record the characters */
  int items = fread (line, 1, count, file);
  if (items != count) {
    printf ("tried to read %d chars for key line, but read %d\n", count, items);
    return -1;
  }
  line [count] = '\0';
  return count;
}

static void make_alpha (char * buffer, int bytes)
{
  int i;

  for (i = 0; i < bytes; i++) {
 /* avoid 'I' and 'O', which look like 1 and 0 */
    buffer [i] = (b1 (buffer + i) % 24) + 'A';
    if (buffer [i] == 'I')
      buffer [i] = 'Y';
    if (buffer [i] == 'O')
      buffer [i] = 'Z';
  }
}

void random_fill (char * buffer, int bytes, int may_call_random, int alpha)
{
  int i;
  if (may_call_random) {
/* initialize with random () in case /dev/urandom and /dev/random don't work */
    for (i = 0; i < bytes; i++)
       buffer [i] = random ();
  }
  if (alpha) make_alpha (buffer, bytes);

/* now try to read /dev/urandom (preferred), and if that fails, /dev/random */
  int fd = open ("/dev/urandom", O_RDONLY);
  if (fd < 0) {
    perror ("open /dev/urandom");
    fd = open ("/dev/random", O_RDONLY);
    if (fd < 0) {
      perror ("open /dev/random");
/* only wrote to the buffer if may_call_random was true */
      return;
    }
  }
  /* read the bytes and store them */
  for (i = 0; i < bytes; i++) {
    if (read (fd, buffer + i, 1) != 1) {
      perror ("read random");
      close (fd);
      if (alpha) make_alpha (buffer, i);
/* wrote i random bytes to the buffer */
      return;
    }
  }
  close (fd);  /* done */
  if (alpha) make_alpha (buffer, bytes);
}

static char * temp_filename (int is_private, char * contacts_dir)
{
#define TEMP_PRIV	"new_private_key"
#define TEMP_PUB	"new_public_key"
  /* use a statically allocated array, so it doesn't have to be free'd */
  static int allocated = 0;
  static char * fname = NULL;
  int maxstrlen = 0;
  if (is_private)
    maxstrlen = strlen (contacts_dir) + 1 + strlen (TEMP_PRIV) + 1;
  else
    maxstrlen = strlen (contacts_dir) + 1 + strlen (TEMP_PUB ) + 1;
  if (maxstrlen > allocated) {
    fname = realloc (fname, maxstrlen);
    if (fname == NULL) {
      printf ("unable to allocate file name of %d characters\n", maxstrlen);
      return 0;
    }
    allocated = maxstrlen;
  }
  if (is_private)
    snprintf (fname, maxstrlen, "%s/%s", contacts_dir, TEMP_PRIV);
  else
    snprintf (fname, maxstrlen, "%s/%s", contacts_dir, TEMP_PUB);
  return fname;
#undef TEMP_PRIV
#undef TEMP_PUB
}

/* returns the number of bytes saved if it succeeds, -1 if it does not */
/* if key_bytes is NULL, simply returns the number of bytes needed */
static int get_key_bytes (RSA * key, char * key_bytes, int ksize)
{
  if (key == NULL)
    return 0;
  int bn_size = BN_num_bytes (key->n);
  if ((ksize == 0) || (key_bytes == NULL))
    return bn_size;
  char * buffer = malloc (bn_size);
  if (buffer == NULL) {
    printf ("get_key_bytes unable to allocate bignum buffer\n");
    return -1;
  }
  BN_bn2bin (key->n, buffer);
  int bytes = bn_size;
  if (bytes > ksize)
    bytes = ksize;
  memcpy (key_bytes, buffer, bytes);
  free (buffer);
  if (bytes < ksize)
    memset (key_bytes + bytes, 0, ksize - bytes);
/*
  printf ("get_key_bytes filled in %d bytes, first one is %02x\n", bytes,
          b1 (key_bytes));
*/
  return bytes;  /* how many bytes were actually filled in */
}

/* keys are stored as the PEM-encoded key.
 * Returns the number of bytes read for success, 0 for fail */
static int get_key (char * fname, RSA ** key, int is_private)
{
  off_t filesize;
  FILE * filep;
  int fd = open_create_file (fname, &filesize, &filep);
  if (filesize == 0) {
    printf ("unable to find %skey file %s\n",
           (is_private) ? "private " : "", fname);
    char * x = NULL;
    *x = '\0';
    return 0;   /* failed */
  }

  if (is_private) {
    /* if all goes well, creates new RSA */
    *key = PEM_read_RSAPrivateKey (filep, NULL, NULL, NULL);
  } else {
    *key = PEM_read_RSAPublicKey (filep, NULL, NULL, NULL);
  }
  if (*key == NULL) {
    printf ("error reading %skey from file %s\n",
            (is_private) ? "private " : "", fname);
    return 0;
  }

  long finish = ftell (filep);
  if (finish != filesize)
    printf ("only read %ld bytes from key file with %ld bytes\n",
            finish, filesize);
  fclose (filep);   /* this should also close fd */
  return finish;
}

/* gets the private key and public key and name.
 * return -1 in case of errors, and otherwise, the number of bytes read.
 */
static int get_keys (FILE * file, RSA ** priv_key, RSA ** pub_key)
{
  long start = ftell (file);

#if 0   /* no longer storing names in key files */
  if (my_readline (file, name) <= 0) {
    /* eof, do not print: printf ("error reading key name from file\n"); */
    return -1;
  }
#endif /* 0 */
  if (priv_key != NULL) {
    /* if all goes well, creates new RSA */
    *priv_key = PEM_read_RSAPrivateKey (file, NULL, NULL, NULL);
    if (*priv_key == NULL) {
      printf ("error reading private key from file\n");
      return -1;
    }
  }
  if (pub_key != NULL) {
    *pub_key = PEM_read_RSAPublicKey (file, NULL, NULL, NULL);
    if (*pub_key == NULL) {
      printf ("error reading public key from file\n");
      if (priv_key != NULL)
        RSA_free (*priv_key);
      return -1;
    }
  }

  long finish = ftell (file);
  return (finish - start);
}

/* allocates the buffer needed to hold the bytes in the file +1
 * reads the file contents into the buffer, appending '\0'
 * sets value to point to the newly allocated buffer.
 * returns the number of bytes in case of success, or 0 for failure.
 */
static int get_file_contents (char * fname, char ** value)
{
  off_t filesize;
  int fd = open_create_file (fname, &filesize, NULL);
  if (fd < 0) {
    printf ("unable to open file %s to read content\n", fname);
    return 0;
  }
  if (filesize == 0) {
    printf ("file %s is empty\n", fname);
    return 0;
  }
  char * buf = malloc (filesize + 1);
  if (buf == NULL) {
    printf ("unable to allocated %ld bytes to read file %s\n",
            filesize + 1, fname);
    close (fd);
    return 0;
  }
  int bytes = read (fd, buf, filesize);
  if (bytes != filesize) {
    perror ("read/content");
    printf ("tried to read %ld bytes from %s, read %d\n",
            filesize, fname, bytes);
    close (fd);
    free (buf);
    return 0;
  }
  close (fd);
  buf [filesize] = '\0';
  *value = buf;
  return filesize + 1;
}

/* returns 1 if successful, 0 otherwise */
static int save_key_file (char * fname, RSA * key, int is_private)
{
  FILE * filep;
  off_t filesize;
  int fd = open_create_file (fname, &filesize, &filep);
  if (fd < 0) {
    printf ("unable to open file %s to save %s key\n", fname,
            (is_private) ? "private" : "public");
    return 0;
  }
  if (filesize > 0)
    printf ("note: saving key to file %s which is not empty (%ld bytes)\n",
            fname, filesize);
/*
  char fourb [4];
  get_key_bytes (key, fourb, 4);
  printf ("saving %s key, first four bytes are %d %d %d %d\n",
          (is_private) ? "private" : "public",
          fourb [0] & 0xff, fourb [1] & 0xff,
          fourb [2] & 0xff, fourb [3] & 0xff);
*/

  if (is_private) {
    if (! PEM_write_RSAPrivateKey (filep, key, NULL, NULL, 0, NULL, NULL)) {
      printf ("unable to write private key\n");
      return 0;
    }
    /* printf ("saved private key\n"); */
  } else {
    if (! PEM_write_RSAPublicKey (filep, key)) {
      printf ("unable to write public key\n");
      return 0;
    }
  }

  fclose (filep);   /* should close fd as well */
  return 1;
}

/* returns 1 if successful, 0 otherwise */
static int save_file_contents (char * fname, char * value, int bytes)
{
  off_t filesize;
  int fd = open_create_file (fname, &filesize, NULL);
  if (fd < 0) {
    printf ("unable to open file %s to save content %s\n", fname, value);
    return 0;
  }
  if (filesize > 0)
    printf ("note: saving value to file %s which is not empty (%ld bytes)\n",
            fname, filesize);

  int written = write (fd, value, bytes);
  if (written != bytes) {
    perror ("write/content");
    printf ("tried to write %d bytes, wrote %d\n", bytes, written);
    close (fd);
    return 0;
  }
  close (fd);
  return 1;
}

/* saves the name, the private (if supplied) key, and the public key.
 * return -1 in case of errors, and otherwise, the number of bytes written.
 */
static int save_key (FILE * file, RSA * priv_key, RSA * pub_key, char * name)
{
  long start = ftell (file);
  fprintf (file, "%s\n", name);
  if (priv_key != NULL) {
    if (! PEM_write_RSAPrivateKey (file, priv_key,
                                   NULL, NULL, 0, NULL, NULL)) {
      printf ("unable to write private key\n");
      return -1;
    }
    printf ("saved private key\n");
  }
  if (! PEM_write_RSAPublicKey (file, pub_key)) {
    printf ("unable to write public key\n");
    return -1;
  }

  long finish = ftell (file);
  return (finish - start);
}

/* returns 1 if the two numbers are the same, 0 otherwise */
static int same_bn (BIGNUM * n1, BIGNUM * n2)
{
  if ((n1 == NULL) && (n2 == NULL))
    return 1;
  if ((n1 == NULL) || (n2 == NULL))
    return 0;
  int result = BN_cmp (n1, n2);
  if (result == 0)
    return 1;
  return 0;
}

/* returns 1 if the public part of the two keys matches, 0 otherwise */
static int same_key (RSA * k1, RSA * k2)
{
  return (same_bn (k1->n, k2->n) && same_bn (k1->e, k2->e));
}

/* on each call, initialize the random number generator, needed by
 * RSA_public_encrypt */
static void init_random_generator ()
{
  unsigned int random_seed = time (NULL);
  /* 0: do not call "random" in trying to pick a random seed. */
  random_fill ((char *) (&random_seed), sizeof (random_seed), 0, 0);
  srandom (random_seed);
}

void callback (int type, int count, void * arg)
{
  if (type == 0)
    printf (".");
  if (type == 1)
    printf (",");
  if (type == 2)
    printf ("!");
}

#define RSA_E_VALUE	65537
/* called if we have no private keys */
static RSA * generate_private_key ()
{
  int bits = 4096;
  printf ("generating %d-bit private key", bits);
  RSA * result = RSA_generate_key (bits, RSA_E_VALUE, callback, NULL);
  printf ("\n");

  /* debugging
  int n_size = BN_num_bytes (result->n);
  int e_size = BN_num_bytes (result->e);
  printf ("in generated key, n has size %d, e has size %d\n", n_size, e_size); 
  char buffer [1000];
  BN_bn2bin (result->n, buffer);
  printf ("n is: ");
  print_buffer (buffer, n_size, 1);
  BN_bn2bin (result->e, buffer);
  printf ("e is: ");
  print_buffer (buffer, e_size, 1);
*/

  return result;
}

/* there is a separate set of keys for each contact in the chat program */
/* each set of keys consists of my private and public keys, and the contact's
 * public key, saved in files 'priv', 'pub', and 'contact'.
 * The contact's name is saved in file 'name'
 * All other files in each contact's directory are content files, to be
 * managed by the application.
 */

struct key_info {
  char * name;
  char * path;
  RSA * contact_pubkey;
  RSA * pubkey;
  char pubkey_bytes [4];  /* useful sometimes */
  RSA * privkey;
};

#define MAX_KEYS	10000
static struct key_info keys [MAX_KEYS];
static int saved_keys = 0;
char * contacts_dir = NULL;

/* returns the number n of contacts if successful, otherwise -1.
 *
 * if no keys were found, returns 0, but can use
 * keys if they are added with add_key.
 *
 * valid key IDs have values from 1..n */
int init_key_files (char * path, char * program_name)
{
#define CONTACTS_SUBDIR	"contacts"
#define CONTACT_DATE_SAMPLE	"19700101-000000"
/* file name is, e.g. path/program_name/CONTACTS_SUBDIR/CONTACT_DATE/fname */
#define LONGEST_NAME_SAMPLE	CONTACT_DATE_SAMPLE  /* start date of data in the file */
  int maxstrlen = strlen (path) + 1 +
                  strlen (program_name) + 1 +
                  strlen (CONTACTS_SUBDIR) + 1 +
                  strlen (CONTACT_DATE_SAMPLE) + 1 +
                  strlen (LONGEST_NAME_SAMPLE) + 1;
  char * fname = malloc (maxstrlen);
  if (fname == NULL) {
    printf ("unable to allocate file name of %d characters\n", maxstrlen);
    return -1;
  }
  snprintf (fname, maxstrlen, "%s/%s/%s",
            path, program_name, CONTACTS_SUBDIR);
  DIR * dir = opendir (fname);
  if (dir == NULL) {
    perror ("opendir");
    printf ("unable to open directory %s\n", fname);
    free (fname);
    return -1;
  }
  contacts_dir = malloc_strcpy (fname);  /* save the path */
  int key_index = 0;
  while (1) {
    struct dirent * dep = readdir (dir);
    if (dep == NULL)  /* finished */
      break;
    if (key_index >= MAX_KEYS) {
      printf ("more than %d keys!!! aborting\n", MAX_KEYS);
      exit (1);
      break;   /* if we don't exit, end the loop, no point continuing */
    }
    if ((dep->d_name [0] != '.') &&   /* ignore "hidden" files */
        (dep->d_type == DT_DIR)) {    /* only look at subdirectories */
      keys [key_index].path = malloc (maxstrlen);
      snprintf (keys [key_index].path, maxstrlen, "%s/%s/%s/%s/",
                path, program_name, CONTACTS_SUBDIR, dep->d_name);
/* get the name of the contact */
      FILE * filep;
      off_t filesize;
      snprintf (fname, maxstrlen, "%s/%s/%s/%s/%s",
                path, program_name, CONTACTS_SUBDIR, dep->d_name, "name");
      if (! get_file_contents (fname, &(keys [key_index].name)))
        continue;
/* get the private key */
      snprintf (fname, maxstrlen, "%s/%s/%s/%s/%s",
                path, program_name, CONTACTS_SUBDIR, dep->d_name, "priv");
      if (get_key (fname, &(keys [key_index].privkey), 1) <= 0)
        continue;
/* get our public key */
      snprintf (fname, maxstrlen, "%s/%s/%s/%s/%s",
                path, program_name, CONTACTS_SUBDIR, dep->d_name, "pub");
      if (get_key (fname, &(keys [key_index].pubkey), 0) <= 0)
        continue;
      get_key_bytes (keys [key_index].pubkey,
                     keys [key_index].pubkey_bytes,
                     sizeof (keys [key_index].pubkey_bytes));
/* get the contact's public key */
      snprintf (fname, maxstrlen, "%s/%s/%s/%s/%s",
                path, program_name, CONTACTS_SUBDIR, dep->d_name, "contact");
      if (get_key (fname, &(keys [key_index].contact_pubkey), 0) <= 0)
        continue;
      printf ("found key for contact %s\n", keys [key_index].name);
/* done, start next loop */
      key_index++;
    }
  }
  closedir (dir);
  free (fname);
  saved_keys = key_index;
  return key_index;
}

#if 0
/* returns the number n of public keys if successful, otherwise -1.
 * key IDs normally have values from 1..n */
int init_key_file (char * fname, char * myname)
{
  init_random_generator ();
#define PRIV_STRING	".private"
  int priv_size = strlen (ALLNET_API_DIRECTORY) + strlen (fname)
                + strlen (PRIV_STRING) + 1;
  char * priv_name = malloc (priv_size);
  if (priv_name == NULL) {
    printf ("unable to allocate name for private key file\n");
    exit (1);
  }
/*
  snprintf (priv_name, priv_size, "%s%s%s",
            ALLNET_API_DIRECTORY, fname, PRIV_STRING);
*/
  snprintf (priv_name, priv_size, "%s%s", fname, PRIV_STRING);
#undef PRIV_STRING
  off_t private_size, public_size;
  private_fd = open_create_file (priv_name, &private_size, &private_file);
  if (private_fd < 0)
    return -1;
  public_fd = open_create_file (fname, &public_size, &public_file);
  if (public_fd < 0)
    return -1;

  int private = 1;   /* read private keys first */
  while (1) {
    if (saved_keys >= MAX_KEYS)
      break;
    RSA * pub_key = NULL;
    RSA * priv_key = NULL;
    char * name = NULL;
    int b;
    if (private)
      b = get_keys (private_file, &priv_key, &pub_key, &name);
    else
      b = get_keys (public_file, NULL, &pub_key, &name);
    if (b <= 0) {
      if (! private)
        break;   /* normal end of the loop -- finished reading public keys */
      if (num_private_keys == 0) {
        printf ("no key found, generating private key for name '%s'\n", myname);
        name = myname;
        RSA * key = generate_private_key ();
        save_key (private_file, key, key, myname);
        long location = ftell (public_file); /* should be 0, but may not be */
        fseek (public_file, 0, SEEK_END);  /* write the public name too */
        save_key (public_file, NULL, key, myname);
        fseek (public_file, location, SEEK_SET);  /* go back */
        fseek (private_file, 0, SEEK_SET);  /* try to read again */
      } else {
        private = 0;  /* start reading public keys */
      }
    } else { /* b > 0 */
      if (private)
        private_size -= b;
      else
        public_size -= b;
      if (private)
        num_private_keys++;
      int name_size = strlen (name);
      printf ("got ");
      if (private)
        printf ("private (%d) and ", RSA_size (priv_key));
      printf ("public (%d bytes) key for %s (%d chars)\n", RSA_size (pub_key),
              name, name_size);
      int already_have = 0;
      if (! private) {  /* public key, we may already have private key */
        int i;
        for (i = 0; i < num_private_keys; i++) {
          if (same_key (pub_key, keys [i].pubkey)) {
            already_have = 1;    /* matching public key */
            printf ("found public key %s matching private key %s\n", name,
                    keys [i].name);
          }
        }
      }
      if (already_have) {        /* do not save this key */
        if (pub_key != NULL)
          RSA_free (pub_key);
        if (priv_key != NULL)
          RSA_free (priv_key);
        continue;                /* start loop again */
      }
      struct key_info * s = keys + saved_keys;
      s->name = name;
      s->pubkey = pub_key;
      get_key_bytes (pub_key, s->pubkey_bytes, sizeof (s->pubkey_bytes));
      s->has_privkey = private;
      s->privkey = NULL;
      if (private)
        s->privkey = priv_key;
      saved_keys++;

      printf ("first 4 bytes are %02x %02x %02x %02x\n",
              b1 (s->pubkey_bytes), b1 (s->pubkey_bytes + 1),
              b1 (s->pubkey_bytes + 2), b1 (s->pubkey_bytes + 3)); 
    }
  }
  printf ("got %d keys, of which %d are my own\n",
          saved_keys, num_private_keys);
  return saved_keys - num_private_keys;
}
#endif /* 0 */

/* returns 1 if it found a name for the given key.  Puts as much of the
 * name as possible into namebuf.  If not found, returns 0, -1 for errors.
 * if a key does not have a name, namebuf is set to "",
 * but name_for_key still returns 1.
 */
int name_for_key (int key_id, char * namebuf, int bsize,
                  char * path, int psize)
{
  if ((key_id >= 0) && (key_id < saved_keys)) {
    strncpy (namebuf, keys [key_id].name, bsize - 1);
    namebuf [bsize - 1] = '\0';
    if ((psize > 0) && (path != NULL)) {
      if (psize > strlen (keys [key_id].path))
        strncpy (path, keys [key_id].path, psize);
      else
        strncpy (path, "", psize);
    }
    return 1;
  }
  return 0;
}

/* given any part of a name, returns all the matching key IDs, up to
 * num_keys.  Also returns the total number of matching keys, which
 * may be greater than num_keys.  Returns -1 in case of errors.
 */
int key_for_name (char * name, int * key_id, int num_keys)
{
  int result = 0;
  int i;
  for (i = 0; i < saved_keys; i++) {
    if (strcasestr (keys [i].name, name) != NULL) {
      if (num_keys > result)
        key_id [result] = i;
      result++;
    }
  }
  return result;
}

/* given a key ID, returns the first few bytes of the public key
 * returns the number of bytes returned, never more than num_bytes,
 * or -1 for errors */
int key_bytes (int key_id, char * key_bytes, int num_bytes)
{
  if (num_bytes <= 0)
    return 0;
  if ((key_id >= 0) && (key_id < saved_keys)) {
    if (num_bytes > 4)
      num_bytes = 4;
    memcpy (key_bytes, keys [key_id].pubkey_bytes, num_bytes);
    return num_bytes;
  }
  return 0;
}

/* adds the given key, either temporarily (persist = 0) or permanently
 * (persist != 0, stored in the file).
 * returns the key_id of the new key, or -1 in case of failure.
 */
int add_key (char * name, char * key, int key_len)
{
  /* create a new RSA for the key received from the contact.
   * The man pages discourage this, but I am not sure
   * how else to go about it.  I don't have a DER.
   */
  BIGNUM * n = BN_bin2bn (key, key_len, NULL);
  /* e is always 65537 */
  char e_buf [3] = { 1, 0, 1 };
  BIGNUM * e = BN_bin2bn(e_buf, 3, NULL);
  if ((e == NULL) || (n == NULL)) {
    printf ("unable to create bignums n, e\n");
    return -1;
  }
  RSA * rsa = RSA_new ();
  rsa->n = n;
  rsa->e = e;

  /* see if I already have the key */
  int i;
  for (i = 0; i < saved_keys; i++) {
    if (strcmp (name, keys [i].name) == 0)
      printf ("warning: already have key for %s, new key will be unusable\n",
              name);
    if (same_key (rsa, keys [i].contact_pubkey)) {
      printf ("received key matches existing key %d\n", i);
      return i;
    }
  }

  int maxstrlen = strlen (contacts_dir) + 1 +
                  strlen (CONTACT_DATE_SAMPLE) + 1 +
                  strlen (LONGEST_NAME_SAMPLE) + 1;
  char * fname = malloc (maxstrlen);
  if (fname == NULL) {
    printf ("unable to allocate file name of %d characters\n", maxstrlen);
    return -1;
  }
  time_t now;
  time (&now);
  struct tm * t = gmtime (&now);
  snprintf (fname, maxstrlen, "%s/%04d%02d%02d-%02d%02d%02d",
            contacts_dir, t->tm_year + 1900, t->tm_mon + 1,
            t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec);
  if (mkdir (fname, 0755) != 0) {
    perror ("mkdir");
    printf ("unable to create user directory");
  }

  struct key_info * s = keys + saved_keys;
  s->name = malloc_strcpy (name);
  s->path = malloc_strcpy (fname);
  s->contact_pubkey = rsa;
/*
char debug [4];
get_key_bytes (s->contact_pubkey, debug, 4);
printf ("received key begins with %d %d %d %d\n",
debug [0] & 0xff, debug [1] & 0xff, debug [2] & 0xff, debug [3] & 0xff);
*/
  if (get_key (temp_filename (0, contacts_dir), &(s->pubkey), 0) <= 0) {
    printf ("add_key unable to get new public key %s\n",
            temp_filename (0, contacts_dir));
    return 0;
  }
  get_key_bytes (s->pubkey, s->pubkey_bytes, sizeof (s->pubkey_bytes));
/*
get_key_bytes (s->pubkey, debug, 4);
printf ("public key from %s begins with %d %d %d %d\n",
temp_filename (0, contacts_dir),
debug [0] & 0xff, debug [1] & 0xff, debug [2] & 0xff, debug [3] & 0xff);
*/
  if (get_key (temp_filename (1, contacts_dir), &(s->privkey), 1) <= 0) {
    printf ("add_key unable to get new private key %s\n",
            temp_filename (1, contacts_dir));
    return 0;
  }
/*
get_key_bytes (s->privkey, debug, 4);
printf ("private key from %s begins with %d %d %d %d\n",
temp_filename (1, contacts_dir),
debug [0] & 0xff, debug [1] & 0xff, debug [2] & 0xff, debug [3] & 0xff);
*/

  snprintf (fname, maxstrlen, "%s/name", s->path);
  if (save_file_contents (fname, s->name, strlen (s->name)) <= 0) {
    printf ("add_key unable to save contact name %s in file %s\n",
            s->name, fname);
    return 0;
  }
  snprintf (fname, maxstrlen, "%s/priv", s->path);
  if (! save_key_file (fname, s->privkey, 1)) {
    printf ("add_key unable to save private key in file %s\n", fname);
    return 0;
  }
  snprintf (fname, maxstrlen, "%s/pub", s->path);
  if (! save_key_file (fname, s->pubkey, 0)) {
    printf ("add_key unable to save public key in file %s\n", fname);
    return 0;
  }
  snprintf (fname, maxstrlen, "%s/contact", s->path);
  if (! save_key_file (fname, s->contact_pubkey, 0)) {
    printf ("add_key unable to save contact's public key in file %s\n", fname);
    return 0;
  }

/*
 ============== should unlink at some point =============
  if (unlink (temp_filename (1, contacts_dir)) < 0)
    perror ("unlink private key");
  if (unlink (temp_filename (0, contacts_dir)) < 0)
    perror ("unlink public key");
*/

  int result = saved_keys;
  saved_keys++;
  printf ("add key returning %d\n", result);
  return result;
}

/* creates the new key, and saves it.  The key is not really available
 * until add_key is called
 * returns 1 for success, 0 for failure.
 */
int new_keypair ()
{
  /* generate the key */
  int bits = 4096;
  printf ("generating %d-bit private key", bits);
  RSA * key = RSA_generate_key (bits, RSA_E_VALUE, callback, NULL);
  printf ("\n");

  /* debugging
  int n_size = BN_num_bytes (key->n);
  int e_size = BN_num_bytes (key->e);
  printf ("generated key, n has size %d, e has size %d\n", n_size, e_size); 
  char buffer [1000];
  BN_bn2bin (key->n, buffer);
  printf ("n is: ");
  print_buffer (buffer, n_size, 1);
  BN_bn2bin (key->e, buffer);
  printf ("e is: ");
  print_buffer (buffer, e_size, 1);
*/

  if (! save_key_file (temp_filename (1, contacts_dir), key, 1)) {
    printf ("unable to save private key to file %s\n",
            temp_filename (1, contacts_dir));
    return 0;
  }
  /* printf ("saved private key\n"); */
  if (! save_key_file (temp_filename (0, contacts_dir), key, 0)) {
    printf ("unable to save public key to file %s\n",
            temp_filename (0, contacts_dir));
    return 0;
  }

  return 1;
}


/* fills in cypher with the signed and encrypted text.  key_id is the
 * ID identifying the recipient.  Also fills in the first ksize
 * bytes of the key, padding with zeros if ksize > key length.
 * returns the number of bytes in the cyphertext if successful -- if
 * the value returned > csize, the result does not fit and is not
 * stored in cypher.
 * two bytes are added to the plaintext (before encrypting), the first
 * two bytes of the public key of the key used for signing.
 * returns the total bytes stored in cypher, or
 * returns -1 in case of failure.
 */
int sign_encrypt (char * text, int tsize, char * cypher, int csize,
                  char * owner_id, int key_id, char * key, int ksize)
{
/*
  printf ("sign_encrypt (%s(%d), (%d), %p, %d, (%d))\n",
          text, tsize, csize, owner_id, key_id, ksize);
  print_buffer (owner_id, OWNER_ID_SIZE, 1);
*/
  if ((key_id < 0) || (key_id >= saved_keys)) {
    printf ("invalid key id %d, valid range 0..%d\n", key_id, saved_keys - 1);
    return -1;
  }
  RSA * rsa = keys [key_id].contact_pubkey;   /* their key */

  RSA * signing_rsa = keys [key_id].privkey;  /* my key, for signing */

  if (tsize >= RSA_size (rsa) - 13) {
    /* fix later to use session keys.  This is OK for chat I think */
    printf ("error, message size %d larger than maximum allowed %d\n",
            tsize, RSA_size (rsa) - 13);
    return -1;
  }
/* it's 499 for 4096-bit = 512-byte keys
  printf ("maximum message size %d\n", RSA_size (rsa) - 13); */

  int hash_size = 0;
  if (owner_id != NULL)
    hash_size = OWNER_ID_SIZE;

  int total_bytes = RSA_size (rsa) + RSA_size (signing_rsa);
  if (csize < RSA_size (rsa) + RSA_size (signing_rsa)) {
    printf ("cyphertext size provided %d, but need %d + %d = %d bytes\n",
            csize, RSA_size (rsa), RSA_size (signing_rsa), total_bytes);
    return -1;
  }

  char text_copy [2052 + OWNER_ID_SIZE];
  if (sizeof (text_copy) < tsize + 2 + hash_size) {
    printf ("only set up to encrypt up to %d bytes, %d given, failing\n",
            sizeof (text_copy) - 2 - hash_size, tsize);
    return -1;
  }
  text_copy [0] = keys [key_id].pubkey_bytes [0];
  text_copy [1] = keys [key_id].pubkey_bytes [1];
  if (owner_id != NULL)
    memcpy (text_copy + 2, owner_id, hash_size);
  memcpy (text_copy + 2 + hash_size, text, tsize);

  /* signature follows cyphertext */
  unsigned char * signature = cypher + RSA_size (rsa);
  unsigned int siglen;
  if (RSA_sign (NID_md5, text_copy, 2 + tsize + hash_size,
                signature, &siglen, signing_rsa) != 1) {
    unsigned long e = ERR_get_error ();
    printf ("RSA signature failed %ld/%s\n", e, ERR_error_string (e, NULL));
    return -1;
  }

  int bytes = RSA_public_encrypt (2 + hash_size + tsize, text_copy,
                                  cypher, rsa, RSA_PKCS1_PADDING);
  if (bytes != RSA_size (rsa)) {
    unsigned long e = ERR_get_error ();
    printf ("RSA encryption failed %ld/%s %d\n", e, ERR_error_string (e, NULL),
            bytes);
    return -1;
  }
  if ((key != NULL) && (ksize > 0))
    get_key_bytes (rsa, key, ksize);  /* bytes to send with the packet */
/*
  printf ("  ==> %d: ", total_bytes);
  print_buffer (cypher, 10, 1);
*/
  return total_bytes;
}

/* returns 1 for a possible match, 0 for guaranteed no match */
static int matching_key (RSA * key, char * key_hint, int ksize)
{
  if ((key_hint == NULL) || (ksize <= 0))
    return 1;   /* always match */
  char buffer [sizeof (struct allnet_pubkey_addr)];
  int bsize = sizeof (struct allnet_pubkey_addr);
  if (ksize < bsize)
    bsize = ksize;  /* now ksize >= bsize */
  int bytes = get_key_bytes (key, buffer, bsize);
  if (bytes > bsize) {
    printf ("matching_key error, got %d bytes, expecting at most %d\n",
            bytes, bsize);
    bytes = bsize;
  }
  /* bytes may be less than ksize -- only compare the first bytes */
  int comparison = memcmp (buffer, key_hint, bytes);
/*
  printf ("matching_key (%d: %02x =? %02x) %smatch\n", bytes,
          b1 (buffer), b1 (key_hint), (comparison == 0) ? "" : "no ");
*/
  if (comparison == 0)
    return 1;   /* match */
  return 0;  /* no match */
}

/* if successfully decrypts and verifies using the given key, fills in
 * text (up to tsize) and returns the total number of characters in
 * text (which may exceed tsize).
 *
 * if unable to verify, returns -1
 */
static int do_decrypt_verify (char * cypher, int csize, char * text, int tsize,
                              int key, int * key_id_result)
{
  if ((key < 0) || (key >= saved_keys)) {
    printf ("invalid key id %d, valid range 0..%d\n", key, saved_keys - 1);
    return -1;
  }
  RSA * rsa = keys [key].privkey;
  int encrypted_size = RSA_size (rsa);

  if (encrypted_size >= csize) {
    printf ("unable to decrypt message of size %d, cypher size %d needs %d\n",
            csize, key, encrypted_size);
    return -1;
  }

  char clear [2048 * 4];
  if (sizeof (clear) < encrypted_size) {
    printf ("RSA size %d too large for buffer size %d\n", encrypted_size,
            sizeof (clear));
    return -1;
  }
  int decrypted = RSA_private_decrypt (encrypted_size, cypher, clear,
                                       rsa, RSA_PKCS1_PADDING);
/*
  printf ("decryption with private key %d gave %d\n", key, decrypted);
*/
  if (decrypted < 0)
    return -1;
  if (decrypted < 2) {
    printf ("error: received packet decryption gave %d bytes\n", decrypted);
    return -1;
  }

  /* decryption successful, verify the signature */
  /* try all the public keys for verification */
  int i;
  for (i = 0; i < saved_keys; i++) {
    RSA * signing_rsa = keys [i].contact_pubkey;
    char key_bytes [2];
    if (get_key_bytes (signing_rsa, key_bytes, sizeof (key_bytes)) != 2) {
      printf ("error in getting first two bytes of key %d\n", i);
      continue;   /* invalid key? */
    }
/*
    printf ("key index %d has bytes %d %d, received key bytes %d %d\n",
            i, b1 (key_bytes), b1 (key_bytes + 1), b1 (clear), b1 (clear + 1));
*/
    if ((b1 (key_bytes) == b1 (clear)) &&
        (b1 (key_bytes + 1) == b1 (clear + 1))) { /* matching sig */
      int sign_size = RSA_size (signing_rsa);

      if (encrypted_size + sign_size != csize) {
        printf ("encrypted %d + sign %d != csize %d, ignoring key %d\n",
                encrypted_size, sign_size, csize, i);
        continue;   /* cannot be a match, try the next key */
      }
      char * sigp = cypher + encrypted_size;
      if (RSA_verify (NID_md5, clear, decrypted, sigp,
                      sign_size, signing_rsa)) {
        int bytes = decrypted - 2;
        if (bytes > tsize)
          bytes = tsize;
        memcpy (text, clear + 2, bytes);
        if (key_id_result != NULL)
          *key_id_result = i;
        return decrypted - 2;
      }
      printf ("verification with public key %d failed, bytes are\n", i);
      print_buffer (cypher, csize, 1);
    } else {
/*
      printf ("public key %d does not match first two bytes of message: ", key);
      printf ("%02x:%02x != %02x:%02x\n", b1 (key_bytes), b1 (key_bytes + 1),
              b1 (clear), b1 (clear + 1));
*/
    }
  }
  /* none of the keys verified, return */
  return -1;
}

/* if successfully decrypts and verifies, fills in text (up to tsize) and
 * returns the total number of characters in text (which may exceed tsize).
 * key_hint, if not NULL, is used to restrict the number of personal keys
 * which need to be tried.
 *
 * if unable to find a key to decrypt and verify, returns -1
 */
int decrypt_verify (char * cypher, int csize, char * text, int tsize,
                    char * key_hint, int ksize, int * key_id)
{
  int i;
  for (i = 0; i < saved_keys; i++) {
    if (matching_key (keys [i].pubkey, key_hint, ksize)) {
      int result = do_decrypt_verify (cypher, csize, text, tsize, i, key_id);
/*
      printf ("result of do_decrypt_verify on key %d is %d\n", i, result);
 */
      if (result > 0)
        return result;
    }
  }
  return -1;
}

/* creates a message to be sent as part of a key exchange.
 * string must be shared among the two sides and must be hard for
 * an attacker to guess -- recommend a random string at least 10
 * letters long.
 *
 * storage for the message is allocated dynamically, and the caller
 * must free() the result of this call.  The length of the resulting
 * message is stored in len (the message is not null-terminated).
 *
 * The way this works is:
 * each side chooses a public key K and a nonce N.
 * Both sides must share a string S which is hard to guess.
 *
 * Then, each side computes a hash H = hash (S, K, N), and sends to
 * the peer (H, K, N).  The receiver receiving (H', K', N') can
 * recompute H' = hash (S, K', N').
 *
 * If H equals H', and if S is hard to guess, the receiver can trust
 * that the sender knows S, and that K' equals K and is the correct
 * public key for that sender.
 *
 * note the first 8 bytes of the message are set to KEY_EXCHANGE_BYTES
 */
char * message_for_key_exchange (char * string, int * len)
{
  RSA * pubkey;
  if (get_key (temp_filename (0, contacts_dir), &pubkey, 0) <= 0) {
    printf ("message_for_key_exchange: unable to find new public key %s\n",
            temp_filename (0, contacts_dir));
    return 0;
  }
  int key_len = get_key_bytes (pubkey, NULL, 0);
  printf ("message for key exchange, key is %d bytes\n", key_len);
  int string_len = strlen (string);
#define NONCE_SIZE	64
  int total_size = string_len + key_len + NONCE_SIZE;
  int result_size = CLEAR_ID_SIZE + HASH_SIZE + key_len + NONCE_SIZE;

  /* allocate the string to be hashed, and the buffer for the final result */
  char * to_be_hashed = malloc (total_size);
  if (to_be_hashed == NULL) {
    printf ("error allocating %d bytes for hash string\n", total_size);
    return NULL;
  }
  char * result = malloc (result_size);
  if (result == NULL) {
    printf ("error allocating %d bytes for hash message\n", result_size);
    free (to_be_hashed);
    return NULL;
  }

  /* initialize the string to be hashed */
  char * key_pos = to_be_hashed + string_len;
  char * nonce_pos = key_pos + key_len;
  memcpy (to_be_hashed, string, string_len);
  if (get_key_bytes (pubkey, key_pos, key_len) < 0) {
    printf ("error getting %d bytes for key\n", key_len);
    free (to_be_hashed);
    free (result);
    return NULL;
  }
/*
  printf ("first 4 bytes of key to be hashed are %d %d %d %d\n",
          key_pos [0] & 0xff, key_pos [1] & 0xff,
          key_pos [2] & 0xff, key_pos [3] & 0xff); 
*/
  random_fill (nonce_pos, NONCE_SIZE, 1, 0);

  /* fill in all the message except for the hash */
  memcpy (result, KEY_EXCHANGE_BYTES, CLEAR_ID_SIZE);
  memcpy (result + CLEAR_ID_SIZE + HASH_SIZE, key_pos, key_len + NONCE_SIZE);

  /* compute the hash, storing it near the beginning of the result */
  allnet_hash (to_be_hashed, total_size, result + CLEAR_ID_SIZE);
  free (to_be_hashed);

  *len = result_size;
  return result;   /* caller should free result */

#if 0
  if ((key_id < 0) || (key_id >= saved_keys)) {
    printf ("illegal key ID %d, max is %d\n", key_id, saved_keys - 1);
    return NULL;
  }
  struct key_info * kip = keys + key_id;
  int key_len = get_key_bytes (kip->pubkey, NULL, 0);
  printf ("message for key exchange, key %d is %d bytes\n", key_id, key_len);
  int string_len = strlen (string);
/* #define NONCE_SIZE	64 */
  int total_size = string_len + key_len + NONCE_SIZE;
  int result_size = CLEAR_ID_SIZE + HASH_SIZE + key_len + NONCE_SIZE;

  /* allocate the string to be hashed, and the buffer for the final result */
  char * to_be_hashed = malloc (total_size);
  if (to_be_hashed == NULL) {
    printf ("error allocating %d bytes for hash string\n", total_size);
    return NULL;
  }
  char * result = malloc (result_size);
  if (result == NULL) {
    printf ("error allocating %d bytes for hash message\n", result_size);
    free (to_be_hashed);
    return NULL;
  }

  /* initialize the string to be hashed */
  char * key_pos = to_be_hashed + string_len;
  char * nonce_pos = key_pos + key_len;
  memcpy (to_be_hashed, string, string_len);
  if (get_key_bytes (kip->pubkey, key_pos, key_len) < 0) {
    printf ("error getting %d bytes for key\n", key_len);
    free (to_be_hashed);
    free (result);
    return NULL;
  }
  random_fill (nonce_pos, NONCE_SIZE, 1, 0);

  /* fill in all the message except for the hash */
  memcpy (result, KEY_EXCHANGE_BYTES, CLEAR_ID_SIZE);
  memcpy (result + CLEAR_ID_SIZE + HASH_SIZE, key_pos, key_len + NONCE_SIZE);

  /* compute the hash, storing it near the beginning of the result */
  allnet_hash (to_be_hashed, total_size, result + CLEAR_ID_SIZE);
  free (to_be_hashed);

  *len = result_size;
  return result;   /* caller should free result */
#endif /* 0 */
}

/* verifies whether the received message matches the given string.
 * returns -1 in case of no match, and the new key ID in case of success.
 * returns -2 in case this is my own packet.
 * If the new key matches an existing key, returns the key ID of the
 * existing key.  The key is saved with "key_name" as the name for the key.
 */
int verify_key_exchange (char * msg, int len, char * string, char * key_name)
{
  int key_len = len - (HASH_SIZE + NONCE_SIZE + CLEAR_ID_SIZE);

  /* a few basic checks on message length */
  if (key_len <= 0) {
    printf ("verify_key_exchange, received msg %d, but need %d + key bytes\n",
            len, HASH_SIZE + NONCE_SIZE);
    return -1;
  }
  if ((key_len & (key_len - 1)) != 0) /* key_len is not a power of 2 */
    printf ("verify_key_exchange, computed key_len %d, not a power of two\n",
            key_len);
  if (memcmp (KEY_EXCHANGE_BYTES, msg, CLEAR_ID_SIZE) != 0) {
    char prt [CLEAR_ID_SIZE + 1];
    memcpy (prt, msg, CLEAR_ID_SIZE);
    prt [CLEAR_ID_SIZE] = 0;
    printf ("verify_key_exchange, msg bytes are %s, not key exchange\n", prt);
    return -1;
  }

  char * key = msg + HASH_SIZE + CLEAR_ID_SIZE;
  /* is this my key? */
  RSA * pubkey;
  if (get_key (temp_filename (0, contacts_dir), &pubkey, 0) > 0) {
    char new_keyb [5000];
    if (sizeof (new_keyb) < key_len) {
      printf ("key size is %d, only allocated 5000\n", key_len);
      exit (1);
    }
    int new_key_len = get_key_bytes (pubkey, new_keyb, key_len);
    if (memcmp (new_keyb, key, key_len) == 0) {
      /* printf ("received my own key, ignoring\n"); */
      return -2;
    }
  }

/*
printf ("key for verification begins with %d %d %d %d\n",
key [0] & 0xff, key [1] & 0xff, key [2] & 0xff, key [3] & 0xff);
*/

  /* create the buffer for hashing */
  char * nonce = key + key_len;
  int string_len = strlen (string);
  int total_size = string_len + key_len + NONCE_SIZE;
  char * to_be_hashed = malloc (total_size);
  if (to_be_hashed == NULL) {
    printf ("verify error allocating %d bytes for hash string\n", total_size);
    return -1;
  }
  memcpy (to_be_hashed, string, string_len);
  memcpy (to_be_hashed + string_len, key, key_len);
  memcpy (to_be_hashed + string_len + key_len, nonce, NONCE_SIZE);

  /* compute the hash, storing it in hash */
  char hash [HASH_SIZE];
  allnet_hash (to_be_hashed, total_size, hash);

  /* is it correct? */
  if (memcmp (hash, msg + CLEAR_ID_SIZE, HASH_SIZE) != 0) {
    printf ("received bad hash ");
    print_buffer (msg + CLEAR_ID_SIZE, HASH_SIZE, 1);
    printf ("computed hash ");
    print_buffer (hash, HASH_SIZE, 1);
    return -1;
  }
  /* if we get here, the hash includes the string, public key, and nonce,
   * and so should be correct, with all the caveats of course. */
  /* save the key in the file and we are done */
  return add_key (key_name, key, key_len);
}

/* result is malloc'd, should be free'd */
char * random_string (int chars)
{
  char * result = malloc (chars + 1);
  if (result == NULL) {
    printf ("error allocating %d bytes for hash message\n", chars + 1);
    return NULL;
  }
  int i;
  for (i = 0; i < chars; i++)
    result [i] = 'a' + (random () % 26);  /* only use lowercase letters */
  result [chars] = '\0';
  return result;
}

/* do the contents of the buffer contain the hash and owner ID, and
 * does the owner ID correctly hash to the hash value?
 */
int is_valid_ack (char * buffer, int bsize, char * owner_id, int osize)
{
/*
  printf ("is_valid_ack, %d bytes are\n", osize);
  print_buffer (owner_id, osize, 1);
  print_buffer (buffer, osize, 1);
*/
  if (bsize < CLEAR_ID_SIZE + osize)
    return 0;
  if (memcmp (buffer, ACKNOWLEDGEMENT_BYTES, CLEAR_ID_SIZE) != 0)
    return 0;
  char * received_id = buffer + CLEAR_ID_SIZE;
  if (memcmp (received_id, owner_id, osize) != 0)
    return 0;
  return 1;
}

/* result is malloc'd, should be free'd */
char * acknowledgement_message ()
{
}
