/* creates, parses, and prints hello messages. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <mhash.h>

#include "types.h"
#include "messages.h"

static int round_up (int n, int to)
{
  if ((n % to) == 0)
    return n;
  /* if n is 14 and to is 8, should return 16: 14 + (8 - 6) = 15 + 2 */
  return (n + (to - (n % to)));
}

int header_size (struct allnet_addr * route, int num_addrs, int osize)
{
  int offset = 4 + 4 + 8;  /* 2 shorts, 3 chars+pad, 8 bytes of length */
  if (num_addrs == 0)
    return round_up (offset + osize, 8);   /* no addresses */

  int i;
  for (i = 0; i < num_addrs; i++) {
    char buf [1];
    int size = store_addr (route + i, buf, 0);
    if (size >= 0) {  /* should be negative, since we did not give a buffer */
      printf ("error: unknown address size %d\n", size);
      exit (1);
    }
/*
    printf ("address %d has size %d bytes\n", i, -size);
*/
    offset += (- size);
  }
/*
  printf ("header has size %d bytes\n", offset);
*/
  return round_up (offset + osize, 8);
}

static void store_header (char * buffer, int hsize,
                          allnet_data_len total_length, int type,
                          struct allnet_addr * route, int num_addrs, int used,
                          char * owner_id, int osize)
{
  if (hsize >= ALLNET_HEADER_FIXED_SIZE) {
    bs2 (buffer    , ALLNET_VERSION);
    bs2 (buffer + 2, type);
    bs1 (buffer + 4, num_addrs);
    bs1 (buffer + 5, used);
    bs1 (buffer + 6, osize);
    bs1 (buffer + 7, 0);
    bs8 (buffer + 8, total_length);
  }
  int chars = store_route (route, num_addrs, used,
                           buffer + ALLNET_HEADER_FIXED_SIZE,
                           hsize - ALLNET_HEADER_FIXED_SIZE, 0);
  /* sanity check, should not be needed */
  if (chars <= 0) {
    printf ("error in store_header: %d header, %d owner bytes, %d result\n",
            hsize, osize, chars);
    exit (1);
  }
  if (osize != 0) {
    /* sanity check, should not be needed */
    if (round_up (ALLNET_HEADER_FIXED_SIZE + chars + osize, 8) != hsize) {
      printf ("error in store_header: %d header, %d + %d + %d ++ = %d\n",
              hsize, ALLNET_HEADER_FIXED_SIZE, osize, chars,
              round_up (ALLNET_HEADER_FIXED_SIZE + chars + osize, 8));
      exit (1);
    }
    /* if osize is non-zero, owner_id should not be NULL */
    memcpy (buffer + ALLNET_HEADER_FIXED_SIZE + chars, owner_id, osize);
  }
  int filled = ALLNET_HEADER_FIXED_SIZE + chars + osize;
  if (hsize > filled)
    memset (buffer + filled, 0, hsize - filled);
}

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_hello.  Store_hello also stores the allnet header. */

int store_hello (allnet_services my_services,
              /* if num_peers is zero, peers may be NULL */
                 int num_peers, struct allnet_peer * peers,
                 struct allnet_addr * route, int num_addrs, int used,
                 char * buffer, int bsize)
{
  allnet_data_len total_length = 0;
  int hsize = header_size (route, num_addrs, 0);
  int offset = hsize + 10;
  int bad_result = (offset >= bsize);
  int use_offset = (bad_result) ? 0 : offset;

  if (! bad_result) {
    bs8 (buffer + hsize, my_services);
    bs2 (buffer + hsize + 8, num_peers);
  }

  if (num_peers > 0) {
    int i;
    for (i = 0; i < num_peers; i++) {
      int result = store_peer (peers + i, buffer + use_offset,
                               bsize - use_offset);
      if (result == 0) return 0;
      if (result < 0) { bad_result = 1; result = (-result); }
      offset += result;
      use_offset = (bad_result) ? 0 : offset;
    }
  }
  total_length = offset;
  printf ("store_hello, total length %lld = %d header + %lld data\n",
          total_length, hsize, total_length - hsize);

  if (! bad_result)
    store_header (buffer, hsize, total_length,
                  ALLNET_TYPE_HELLO, route, num_addrs, used, NULL, 0);
  if (bad_result)
    return (- total_length);
  return total_length;
}

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * the result should be of size at least ALLNET_MAX_HEADER_SIZE
 * if successful, the header is stored in *result
 * if successful and osize > 0, owner_info_ptr is set to point to the
 * part of result where the owner info is stored.
 *
 * routes are stored in C format -- that is, every route takes up the
 * same number of bytes.  This is different from the format sent on the
 * network, where routes are stored in as few bytes as possible
 */
int parse_header (char * buffer, int bsize,
                  struct allnet_header * result, char ** owner_info_ptr)
{
  if (bsize < 16) return -1;
  result->allnet_version = b2 (buffer);
  result->allnet_type = b2 (buffer + 2);
  int num_addrs = b1 (buffer + 4);
  int addrs_used = b1 (buffer + 5);
  int osize = b1 (buffer + 6);
  result->allnet_num_addrs = num_addrs;
  result->allnet_addrs_used = addrs_used;
  result->allnet_owner_bytes = osize;
  result->allnet_header_pad = b1 (buffer + 7);  /* or should we set it to 0? */
  result->total_length = b8 (buffer + 8);

  int offset = 16;

  if (num_addrs != 0) {
    int i;
    for (i = 0; i < num_addrs; i++) {
      int p = parse_addr (buffer + offset, bsize - offset, result->route + i);
      if (p < 0) return -1;
      offset += p;
    }
  }
  if (osize != 0) {
    if (offset + osize > bsize)
      return -1;
    if (owner_info_ptr != NULL) {
      *owner_info_ptr = ((char *) (result->route + num_addrs));
      memcpy (*owner_info_ptr, buffer + offset, osize);
    }
    offset += osize;
  }
  return round_up (offset, 8);
}

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, the message is stored in *result;
 */
int parse_hello (char * buffer, int bsize,
                 struct allnet_hello * result, int max_peers)
{
  if (bsize < 10) return -1;
  result->sender_provides = b8 (buffer); 
  result->num_peers = b2 (buffer + 8);
  int offset = 10;

  if (result->num_peers > max_peers) {
    printf ("parsing hello message with %d peers, but max_peer is %d\n",
            result->num_peers, max_peers);
    return -1;
  }

  int i;
  for (i = 0; i < result->num_peers; i++) {
    int count = parse_peer (buffer + offset, bsize - offset, result->peers + i);
    if (count < 0) return -1;
    offset += count;
  }

  return offset;
}

int hello_to_string (struct allnet_hello * hello,
                     char * string, int ssize)
{
  strncpy (string, "not implemented", ssize);
}

void print_hello (struct allnet_hello * hello, int eol)
{
  printf ("hello: services %16llx", hello->sender_provides);
  if (eol)
    printf ("\n");
  else
    printf (", ");
  printf ("%d peers: ", hello->num_peers);
  if (eol >= 1) printf ("\n");
  if (hello->num_peers > 0) {
    int i;
    for (i = 0; i < hello->num_peers; i++) {
      print_peer (hello->peers + i, eol);
      if ((eol < 1) && (i + 1 < hello->num_peers)) printf (", ");
    }
  }
}

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_data.  Store_data also stores the allnet header. */
int store_data (char * data, allnet_data_len length, int clear,
                struct allnet_addr * route, int num_addrs, int used,
                char * owner_id, int osize,
                char * buffer, int bsize)
{
  int hsize = header_size (route, num_addrs, osize);
  if (hsize > bsize) {
    return - (hsize + length);
  }
  int type = ALLNET_TYPE_DATA;
  if (clear)
    type = ALLNET_TYPE_CLEAR;
  allnet_data_len total_length = hsize + length;

  store_header (buffer, hsize, total_length, type, route, num_addrs, used,
                owner_id, osize);
  if (total_length <= bsize) {
    memcpy (buffer + hsize, data, length);
    return total_length;
  }
  return -total_length;
}

int data_to_string (allnet_data data, allnet_data_len len,
                    char * string, int ssize)
{
  int result = snprintf (string, ssize, "%lld bytes of data: ", len);
  result += len * 4 + 1;
  if (result > ssize)
    return ( - result);
  int i;
  for (i = 0; i < len; i++) {
    if (i + 1 < len)
      snprintf (string + result, ssize - result, "%02x, ", b1 (data + i));
    else
      snprintf (string + result, ssize - result, "%02x", b1 (data + i));
  }
  return result;
}

void print_data_message (allnet_data data, allnet_data_len len, int eol)
{
  printf ("%lld bytes of data:", len);
  if (eol >= 1)
    printf ("\n");
  else
    printf (" ");
  int i;
  for (i = 0; i < len; i++) {
    printf ("%02x", (data [i] & 0xff));
    if (i + 1 < len) printf (", ");
  }
  if (eol >= 1)
    printf ("\n");
}

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_signed.  Store_signed also stores the allnet header. */

int store_signed (char * data, allnet_data_len length,
                  int sig_type,
                  char * sig, allnet_data_len sig_length,
                  struct allnet_addr * route, int num_addrs, int used,
                  char * owner_id, int osize, char * buffer, int bsize)
{
  allnet_data_len total_length = 0;
  int hsize = header_size (route, num_addrs, osize);
  store_header (buffer, hsize, total_length, ALLNET_TYPE_SIGNED,
                route, num_addrs, used, owner_id, osize);
}

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, pointers to the part of the buffer containing the data
 * and signature, and their lengths, are stored in the last 4 parameters.
 */
int parse_signed (char * buffer, int bsize, int * sig_type,
                  allnet_data_len * data_len, char ** data,
                  allnet_data_len * sig_len, char ** sig)
{
  if (bsize < 8)
    return -1;
  *data_len = b8 (buffer);
  int offset = 8;

  if (bsize < offset + *data_len)
    return -1;
  *data = buffer + offset;
  offset += *data_len;

  if (bsize < offset + 4)
    return -1;
  *sig_type = b4 (buffer + offset);
  offset += 4;

  if (bsize < offset + 8)
    return -1;
  *sig_len = b8 (buffer + offset);
  offset += 8;
  
  if (bsize < offset + *sig_len)
    return -1;
  *sig = buffer + offset;
  offset += *sig_len;

  return offset;
}

int signed_to_string (struct allnet_signed * s, char * string, int ssize)
{
}

void print_signed (struct allnet_signed * s, int eol)
{
  print_data_message (s->message, s->msg_len, eol);
  printf ("signature type %d, ", s->sig_type);
  print_data_message (s->signature, s->sig_len, eol);
}

void print_buffer (char * buffer, int size, int print_eol)
{
  while (size-- > 0) {
    printf ("%02x, ", (*buffer) & 0xff);
    buffer++;
  }
  if (size >= 0)
    printf ("%02x", (*buffer) & 0xff);
  if (print_eol)
    printf ("\n");
}




