/* messages.h: format for messages sent on the allnet network */

#ifndef ALLNET_MESSAGES_H
#define ALLNET_MESSAGES_H

#include "types.h"

#define ALLNET_PORT 	     (htons (0xa119))  /* ALLNet */
#define ALLNET_LOCAL_PORT    (htons (0xa11e))  /* ALLnEt */
#define ALLNET_WIFI_PROTOCOL (htons (0xa119))  /* ALLNet */

/* allnet types, used as the second field of the allnet packet header */
#define ALLNET_TYPE_NONE		0x00  /* no message, only header */
#define ALLNET_TYPE_HELLO		0x01
#define ALLNET_TYPE_DATA		0x03
#define ALLNET_TYPE_CLEAR		0x04  /* cleartext, no signature */
#define ALLNET_TYPE_SIGNED		0x05  /* cleartext with signature */

/* priorities for messages */
#define MAX_PRIORITY			1000000  /* highest priority 1M */
#define LOCAL_MSG_PRIORITY		 900000  /* near the top */
#define FRIEND_PRIORITY			 500000  /* priority for my friends */
#define FRIEND2_PRIORITY		 400000  /* friends of friends */
#define FRIEND3_PRIORITY		 300000  /* f^3 */
#define FRIEND4_PRIORITY		 250000  /* f^4 */
#define FRIEND5_PRIORITY		 200000  /* f^5 */
#define FRIEND6_PRIORITY		 150000  /* f^6 */
#define OTHERS_PRIORITY			  50000  /* f^n, n > 6 */
#define HELLO_PRIORITY			 100000  /* fairly low priority */
#define MIN_PRIORITY			      0  /* no priority */

/* don't send hello packets to a given peer any more than once every
 * 30 minutes */
/* #define HELLO_INTERVAL			(30 * 60)  /* at least 30 minutes */
#define HELLO_INTERVAL			60  /* for testing, one minute */

/* all allnet messages begin with five 2-byte and 1-byte fields: the version
 * number, the allnet type, the number of addresses in the header,
 * the number of addresses in the header that have already been used,
 * the number of bytes used to identify the owner (if any), and a
 * byte that is unused so far.
 * Following these fields is 8 bytes of the total length of the message,
 * followed by the number of addresses specified, and the number
 * of owner bytes specified.
 *
 * after the addresses, there are 0-7 bytes of padding to make sure
 * the entire header is a multiple of 8 bytes.  The padding should be
 * sent as 0s and ignored on receipt.
 *
 * so the minimum header size is 16 -- the 6 fixed fields and the length,
 * and that is only if num_addrs is zero.
 *
 * if we want to be able to construct a reverse path, then "addrs_used"
 * will be at least 1, and "route [0]" will be the address of the
 * original sender.
 * The packet will only be forwarded
 *   if "allnet_addrs_used" <= "allnet_num_addrs", 
 * but if "addrs_used" == "num_addrs", it is only forwarded if there
 *   are hops remaining
 */

#define ALLNET_VERSION     1
struct allnet_header {
  short allnet_version;
  short allnet_type;
  char allnet_num_addrs;          /* number of addresses in the route */
  char allnet_addrs_used;         /* how many addrs have been used */
  char allnet_owner_bytes;        /* number (0..255) of bytes of owner info */
  char allnet_header_pad;         /* unused, 1 byte for alignment */
  allnet_data_len total_length;   /* total length, including this header */
  struct allnet_addr route [0];   /* route for this packet, 0 or more addrs */
  char owner_info [0];     /* bytes of owner info, if any */
  /* padding of 0s as needed to extend the header to a multiple of 8 bytes */
};

/* same as the above, but using the maximum size route and owner info.
 * useful for allocating variables and calling sizeof */
#define ALLNET_HEADER_FIXED_SIZE	(sizeof (struct allnet_header)) /*16*/
struct allnet_max_header {
  char allnet_fixed [ALLNET_HEADER_FIXED_SIZE];
  struct allnet_addr route [255]; /* route for this packet, 0 or more addrs */
  char owner_info [255];   /* bytes of owner info, if any */
  char allnet_padding [2];        /* make the size a multiple of 8 bytes */
};
#define ALLNET_MAX_HEADER_SIZE	(sizeof (allnet_max_header))

/* given a route, how many bytes does the header need? */
extern int header_size (struct allnet_addr * route, int num_addrs, int osize);

/* what follows are the definitions of the message bodies.
 */

/* a hello message is normally sent on a directly connected wireless link,
 * or on an IP network.
 * On an IP network it will have an IP sender address, and a 2-hop route
 * which includes the sender and receiver.
 * On a directly connected wireless link it will have a key or broadcast
 * address as the sender.  The route may start out as an n-hop broadcast
 * address, and be built up into a return route including a number of
 * key addresses, one per hop (perhaps with some intermediate broadcast
 * segments).
 * Either kind of message may have any number of peers. */
/* a recipient of a hello message may or may not reply with the same
 * kind of message, but in any case should keep track of senders and not send
 * a hello message to the same peer more than once every half hour or so,
 * i.e. HELLO_INTERVAL.  For a simple implementation, it can just send
 * hello messages once every half hour, or it can be more specific.
 */
struct allnet_hello {
  /* my services I want to tell you about */
  allnet_services sender_provides;
  /* zero or more other peers I want to tell you about */
  short num_peers;
  struct allnet_peer peers [0];
};

/* a data message with encrypted contents.  Hopefully, the recipient
 * knows what to do with it.  The length is taken from the header,
 * and is usually greater than 0.
 * normally, the first two bytes of the encrypted message are the first
 * two bytes of the sender's public key.
 *
 * to prove ownership of a message, the first 20 bytes might be a hash
 * of something encrypted within the message.  Producing a value that
 * hashes to the first 20 bytes of a message generally means the sender
 * has the authority to delete or otherwise dispose of the message. 
 */
typedef char allnet_data [0];

#define OWNER_ID_SIZE	20		/* first 20 bytes of SHA 512 of
                                           first 20 bytes of the packet */

/* a message with clear contents.  Generally, the first 8 bytes of
 * the message are a string showing the intended use of the message,
 * but this is not required.
 *
 * note the actual structure is identical to allnet_data.
 */
#define CLEAR_ID_SIZE		8
/* a message used in key exchanges */
/* the message following this header contains a 20-byte hash H,
 * a key K, and a 32-byte nonce N */
#define KEY_EXCHANGE_BYTES	"keyxchng"
/* an ack message to confirm that a message has been received.
 * The ack contains the 20-byte secret in the first 20 bytes of the
 * messsage.  This secret is originally only known to the original sender
 * and to the receiver(s) in possession of the private key
 * corresponding to the public key used to encrypt the message.
 *
 * For a message with owner bytes, the ack contents should hash to the
 * owner bytes of the original packet.  This can be used to confirm
 * receipt to anyone in possession of the owner bytes, not just those with
 * knowledge of the message contents, and is usually used to tell those
 * storing the packet that the packet has been received.
 */
#define ACKNOWLEDGEMENT_BYTES	"received"
/* a message to request saved messages.  The original (source) address
 * identifies which messages are requested.  If the source address is
 * not specific (e.g. a broadcast address), all messages are requested.
 * the message only holds the CLEAR_ID REQUEST_BYTES
 */
#define REQUEST_BYTES		"requestd"

/* a message from someone claiming some externally visible ID. */
struct allnet_signed {
  allnet_data_len msg_len;
  allnet_data message;          /* in the clear */
  int sig_type;                 /* not yet defined */
  allnet_data_len sig_len;
  allnet_data signature;        /* sig and certificate and anything else */
};

/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * the result should be of size at least ALLNET_MAX_HEADER_SIZE
 * if successful, the header is stored in *result
 * if successful and owner_bytes > 0, owner_info_ptr is set to point to the
 * part of result where the owner info is stored.
 *
 * routes are stored in C format -- that is, every route takes up the
 * same number of bytes.  This is different from the format sent on the
 * network, where routes are stored in as few bytes as possible
 */
extern int parse_header (char * buffer, int bsize,
                         struct allnet_header * result, char ** owner_info_ptr);

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_hello.  Store_hello also stores the allnet header, parse_hello
 * parses a buffer that does not contain the header. */

extern int store_hello (allnet_services my_services,
                     /* if num_peers is zero, peers may be NULL */
                        int num_peers, struct allnet_peer * peers,
                        struct allnet_addr * route, int num_addrs, int used,
                        char * buffer, int bsize);
extern int parse_hello (char * buffer, int bsize,
                        struct allnet_hello * result, int max_peers);
extern int hello_to_string (struct allnet_hello * hello,
                            char * string, int ssize);
extern void print_hello (struct allnet_hello * hello, int eol);

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_data.  Store_data also stores the allnet header. */

extern int store_data (char * data, allnet_data_len length, int clear,
                       struct allnet_addr * route, int num_addrs, int used,
                       char * owner_id, int osize,
                       char * buffer, int bsize);
extern int data_to_string (allnet_data data, allnet_data_len len,
                           char * string, int ssize);
extern void print_data_message (allnet_data data, allnet_data_len len, int eol);

/* the following four functions implemented in messages.c.  The definitions
 * match the corresponding functions in addr.c, except for the parameters
 * to store_signed.  Store_signed also stores the allnet header. */

extern int store_signed (char * data, allnet_data_len length,
                         int sig_type, char * sig, allnet_data_len sig_length,
                         struct allnet_addr * route, int num_addrs, int used,
                         char * owner_id, int osize, char * buffer, int bsize);
/* returns -1 in case of failure, the number of bytes parsed otherwise.
 * if successful, pointers to the part of the buffer containing the data
 * and signature, and their lengths, are stored in the last 4 parameters.
 */
extern int parse_signed (char * buffer, int bsize, int * sig_type,
                         allnet_data_len * data_len, char ** data,
                         allnet_data_len * sig_len, char ** sig);
extern int signed_to_string (struct allnet_signed * s,
                             char * string, int ssize);
extern void print_signed (struct allnet_signed * s, int eol);

/* utility function in messages.c */
extern void print_buffer (char * buffer, int size, int print_eol);

#endif /* ALLNET_MESSAGES_H */
