/* persist.h: save a limited amount of data persistently, with priorities */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>

#include "types.h"
#include "messages.h"
#include "hash.h"

/* messages are saved in a directory.   Each message has a name
 * consisting of the hash of the owner ID, in hex.
 * For example, 0123456789ABCDEF0123
 *
 * the contents of the message are a 4-byte priority followed by
 * the two 4-byte decrements and two 8-byte times that the
 * priority was last decremented.  This header is
 * followed by the message itself, including the original allnet header
 *
 * all numbers are stored in big-endian byte order.
 */

struct file_header {
  unsigned int priority;
  double second_decrement;
  long long initial_time;
};
/* struct allnet_header ah;  part of the message */

/* when the system has been initialized, cache_dir is not NULL */
static DIR * cache_dir = NULL;
static char * cache_dir_name = NULL;
static long long allowed_storage = 0;
static long long used_storage = 0;
static int minsavedpriority = 0;

static void get_header (char * path,
                        struct file_header * fhp, long long * total_length,
/* if data is not null, it is set to point to newly allocated storage
 * containing the message, but not the file_header */
/* if dest is not null, it is filled with the destination address */
                        char * * data, struct allnet_addr * dest)
{
  if (data != NULL)
    *data = NULL;
  int fd = open (path, O_RDONLY);
  if (fd < 0) {
    perror ("open in get_header");
    exit (1);  /* for now */
    fhp->priority = minsavedpriority;  /* try to delete it */
    return;
  }
  char buffer [sizeof (struct file_header) + sizeof (struct allnet_header)];
  int bsize = sizeof (buffer);
  int bytes = read (fd, buffer, bsize);
  if (bytes != bsize) {
    perror ("read in get_header");
    printf ("tried to read headers of size %d in file %s, got %d bytes\n",
            bsize, path, bytes);
    exit (1);   /* for now */
    close (fd);
    fhp->priority = minsavedpriority;  /* try to delete it */
    return;
  }
  int fhsize = sizeof (struct file_header);
  memcpy ((char *) fhp, buffer, fhsize);
  *total_length = b8 (buffer + fhsize + 8);
  /* recompute the priority based on the time passed since update */
  int original_priority = fhp->priority;
  time_t now = time (NULL);
  double seconds = now - fhp->initial_time;
  if ((fhp->second_decrement > 0) && (seconds > 0)) {
    if (seconds * fhp->second_decrement > fhp->priority)
      fhp->priority = 0;
    else
      fhp->priority -= seconds * fhp->second_decrement;
  }
  printf ("priority %d, after %f seconds (%f/sec) => %d\n",
          original_priority, seconds, fhp->second_decrement, fhp->priority);
  if ((data != NULL) || (dest != NULL)) {  /* need to read the whole message */
    char * buffer = malloc (*total_length);
    if (buffer == NULL) {
      printf ("allocation of %lld bytes failed for persist.c/get_header\n",
              *total_length);
      return;
    }
    if (data != NULL)
      printf ("get_header allocated data buffer %p\n", buffer);
    if (lseek (fd, fhsize, SEEK_SET) == ((off_t) -1)) {
      perror ("lseek");
      free (buffer);
      return;
    }
    bytes = read (fd, buffer, *total_length);
    if (bytes != *total_length) {
      printf ("get_header: unable to read %lld bytes from file %s\n",
              *total_length, path);
      free (buffer);
      return;
    }
    if (dest != NULL) {
      struct allnet_max_header amh;
      struct allnet_header * hp = (struct allnet_header *) (&amh);
      int hsize = parse_header (buffer, *total_length, hp, NULL);
      dest->discrim = ALLNET_NO_ADDR_ID;
      if (b (hp->allnet_num_addrs) > 0)
        *dest = hp->route [b (hp->allnet_num_addrs) - 1];
    }
    if (data == NULL)
      printf ("get_header freeing data buffer %p\n", buffer);
    if (data != NULL)
      *data = buffer;
    else
      free (buffer);
  }
  close (fd);
}

void init_persist (long long max_storage, char * id)
{
  if (cache_dir != NULL) {
    printf ("persistent system already initialized\n");
    return;
  }
  allowed_storage = max_storage;
  used_storage = 0;
  minsavedpriority = 0;
  if (mkdir (id, 0755) < 0) {
    if (errno != EEXIST) {
      perror ("mkdir/cache directory");
      printf ("unable to find or create directory %s\n", id);
      return;
    }
  }
  cache_dir = opendir (id);
  if (cache_dir == NULL) {
    perror ("opendir/message cache");
    printf ("unable to find or create directory %s\n", id);
    return;
  }
  cache_dir_name = id;

  rewinddir (cache_dir);
 /* search through all the messages to compute the minimum priority */
  minsavedpriority = MAX_PRIORITY;
  struct dirent * dep = readdir (cache_dir);
  while (dep != NULL) {
    if ((strcmp (dep->d_name, ".") == 0) ||
        (strcmp (dep->d_name, "..") == 0)) {
      dep = readdir (cache_dir);
      continue;
    }
    char path [1000];
    snprintf (path, sizeof (path) - 1, "%s/%s", cache_dir_name, dep->d_name);
    struct file_header fh;
    long long total_length;
    get_header (path, &fh, &total_length, NULL, NULL);
    if (fh.priority <= minsavedpriority)
      minsavedpriority = fh.priority;
    printf ("used %lld bytes, adding %d + %d + %lld, final size is ",
            used_storage, strlen (path), sizeof (fh), total_length);
    used_storage += strlen (path) + sizeof (fh) + total_length;
    printf ("%lld\n", used_storage);
    dep = readdir (cache_dir);
  }
  printf ("finished initializing cache, %lld bytes of %lld, min priority %d\n",
          used_storage, allowed_storage, minsavedpriority);
}

/* remove files from the cache until the cache size is <= allowed_storage */
static void cleanup ()
{
  printf ("cleaning up cache, initial size is %lld\n", used_storage);
  while (used_storage > allowed_storage) {
    rewinddir (cache_dir);
 /* search through all the messages, deleting those with minsavedpriority,
  * while computing the next minimum priority */
    int new_minpriority = MAX_PRIORITY;
    struct dirent * dep = readdir (cache_dir);
    while (dep != NULL) {
      char path [1000];
      snprintf (path, sizeof (path) - 1, "%s/%s", cache_dir_name, dep->d_name);
      struct file_header fh;
      long long total_length;
      get_header (path, &fh, &total_length, NULL, NULL);
      if (fh.priority <= minsavedpriority) {  /* delete this message */
        printf ("deleting %s with priority\n", path, fh.priority);
        if (unlink (path) != 0) {
          perror ("unlink");
          printf ("unable to delete %s\n", path);
        }
        used_storage -= strlen (path) + sizeof (fh) + total_length;
      } else if (fh.priority < new_minpriority) {
        new_minpriority = fh.priority;
      }
      dep = readdir (cache_dir);
    }
    minsavedpriority = new_minpriority;
  }
}

/* save a message, with the given priority.  A priority of zero doesn't
 * mean the message is deleted immediately, only, that the message has
 * the lowest possible priority.
 *
 * the header is used to determine the size of the message.  The
 * destination is assumed to be the last address in the route of the header.
 *
 * The priority may be set to decrease every second or every 24 hours.
 * if delta_priority_per_second is nonzero, the priority is
 * (logically) decremented every second.
 * if delta_priority_per_day is nonzero, the priority is
 * decremented every 24 hours, after adding back in any decrement
 * corresponding to 86,400 seconds (24 hours)
 */
void save_message (char * msg,
                   struct allnet_header * header, /* parsed from msg */
                   int priority,
                   double delta_priority_per_second)
{
  if (cache_dir == NULL) {
    printf ("save_message: persistent system not initialized\n");
    return;
  }
  int osize = b (header->allnet_owner_bytes);
  if (osize < OWNER_ID_SIZE) { /* do not save messages without owner ID bytes */
    printf ("packet has %d bytes of owner_id, not saving\n", osize);
    return;
  }
  printf ("packet has %d bytes of owner_id, saving\n", osize);
/* assume the longest possible directory has 100 chars, so reserve 500 chars
 * the longest possible hash has 64 hex digits, so reserve 256
 * plus one for the null byte at the end
 */
#define MAX_FNAME_LEN	(500 + 256)
  char new_name [MAX_FNAME_LEN + 1];
  time_t now = time (NULL);
  snprintf (new_name, MAX_FNAME_LEN, "%s/", cache_dir_name);
  char * owner_id = (char *) (header->route + b (header->allnet_num_addrs));
  int len = strlen (new_name);
  int i;
  for (i = 0; i < osize; i++) {
    if (len + 2 >= MAX_FNAME_LEN)
      break;
    snprintf (new_name + len, MAX_FNAME_LEN - len, "%02x", b (owner_id [i]));
    len += 2;
  }
  int fd = open (new_name, O_RDONLY);
  if (fd != -1) {
    printf ("save_message: file %s already exists!\n", new_name);
    close (fd);
    return;
  }
  printf ("save_message: new %s\n", new_name);
  fd = open (new_name, O_RDWR | O_CREAT, 0644);
  if (fd < 0) {
    perror ("open save message");
    return;
  }
  struct file_header fh;
  fh.priority = priority;
  fh.second_decrement = delta_priority_per_second;
  if (fh.second_decrement < 0.0)
    fh.second_decrement = 0.0;  /* cannot increase priority */
  fh.initial_time = now;

  int written1 = write (fd, (char *) (&fh), sizeof (fh));
  int written2 = write (fd, msg, header->total_length);
  if ((written1 != sizeof (fh)) ||
      (written2 != header->total_length)) {
    if (written2 < 0)
      perror ("write save message");
    printf ("save_message: saved %d+%d of %d+%lld bytes\n",
            written1, written2, sizeof (fh), header->total_length);
    if ((written1 < 0) || (written2 < 0))
      exit (1);  /* for now */
  }
  close (fd);
  used_storage += strlen (new_name) + sizeof (fh) + header->total_length;
  if (minsavedpriority > priority)
    minsavedpriority = priority;
  if (used_storage > allowed_storage)
    cleanup ();
}

/* messages matching the first address in the route are returned in
 * order of priority.  Where priority is the same, older messages are
 * returned before newer messages.
 *
 * messages, message_sizes, and priorities should be array of at least
 * max_messages (char *) or ints
 * each entry of messages is set to point to malloc'd storage
 * containing the body of the message, should be free'd.
 *
 * the route is also used to build a valid message that can be sent.
 *
 * returns the number of messages, perhaps greater than max_messages
 * (but only the first max_messages are filled in)
 * may return 0 if there are no messages for this destination
 * will return -1 in case of errors
 */
int get_messages (struct allnet_addr * route, int num_addrs, int used,
                  char ** messages, int * message_sizes, int * priorities,
                  int max_messages)
{
  if (cache_dir == NULL) {
    printf ("get_messsages: persistent system not initialized\n");
    return -1;
  }
  rewinddir (cache_dir);
  int saved_messages = 0;
  int not_saved = 0;
  struct dirent * dep = readdir (cache_dir);
  while (dep != NULL) {
    if ((strcmp (dep->d_name, ".") == 0) ||
        (strcmp (dep->d_name, "..") == 0)) {
      dep = readdir (cache_dir);
      continue;
    }
    char path [1000];
    snprintf (path, sizeof (path) - 1, "%s/%s", cache_dir_name, dep->d_name);
    struct file_header fh;
    long long total_length;
    char * new_message;
    struct allnet_addr destination;
    get_header (path, &fh, &total_length, &new_message, &destination);
/* let ALLNET_NO_ADDR_ID match everything */
    if ((route->discrim != ALLNET_NO_ADDR_ID) &&
        (! matches_addr (&destination, route))) {
      dep = readdir (cache_dir);
/*
      printf ("address ");
      print_addr (route, 0);
      printf (" does not match address ");
      print_addr (&destination, 1);
*/
      printf ("get_messages freeing data buffer %p\n", new_message);
      free (new_message);
      continue;   /* no need to consider this message */
    }
    printf ("address ");
    print_addr (route, 0);
    printf (" matches address ");
    print_addr (&destination, 1);
/* example: new message is lowest priority, saved_messages = max_messages
 * save_index = max_messages - 1.  Assume max_messages > 0, we do not
 * enter loop because of the priority.  do_save = 0, so nothing happens. */
/* same example, but new message is higher priority than last message.
 * fh.priority > priorities [save_index], so enter the else, we
 * discard the last message, decrement saved_messages and save_index,
 * set do_save to true, repeat the loop.
 * now, the while fails because of priority, so we enter the following "if",
 * where we do the assignment and re-increment saved_messages */
    int save_index = saved_messages - 1;
/* only save if have room -- initially if saved < max */
    int do_save = (saved_messages < max_messages);
#define MAX_ARRAY	1000
    if (max_messages > MAX_ARRAY) {
      printf ("persist.c: increase MAX_ARRAY to at least %d\n", max_messages);
      exit (1);
    }
    unsigned long long times [MAX_ARRAY];
    
    while ((save_index >= 0) &&
/* loop while the new message's priority is greater, or
   the priority is the same, and the new start time is less */
           ((fh.priority > priorities [save_index]) ||
            ((fh.priority == priorities [save_index]) &&
             (fh.initial_time < times [save_index])))) {
      if (do_save) {
        messages [save_index + 1] = messages [save_index];
        priorities [save_index + 1] = priorities [save_index];
        message_sizes [save_index + 1] = message_sizes [save_index];
        times [save_index + 1] = times [save_index];
      } else {             /* first loop, discard last entry */
        free (messages [save_index]);
        printf ("get_messages freeing data buffer %p at index %d\n",
                new_message, save_index);
        not_saved++;       /* record that the last entry was discarded */
        saved_messages--;  /* and discard it */
        do_save = 1;       /* discarding one, so go ahead and save new one */
      }
      save_index--;
    }
    if (do_save) {
      messages [save_index + 1] = new_message;
      priorities [save_index + 1] = fh.priority;
      message_sizes [save_index + 1] = total_length;
      times [save_index + 1] = fh.initial_time;
      saved_messages++;
    } else {
      printf ("get_messages freeing new message %p\n", new_message);
      free (new_message);
    }
/* on to the next file */
    dep = readdir (cache_dir);
  }
  return saved_messages + not_saved;
}

/* if this owner ID hashes to this hash, delete any messages with this hash.
 * return the number of messages deleted, which may be 0, or -1 for errors.
 */
int delete_messages (char * owner_id, int osize)
{
  if (cache_dir == NULL) {
    printf ("delete_messsages: persistent system not initialized\n");
    return -1;
  }
  char hash [HASH_SIZE];
  allnet_hash (owner_id, osize, hash);
  char hash_name [MAX_FNAME_LEN + 1];
  int i, j;
  j = 0;
  for (i = 0; i < OWNER_ID_SIZE; i++) {
    if (j + 3 > MAX_FNAME_LEN) {
      break;
    }
    snprintf (hash_name + j, MAX_FNAME_LEN - j, "%02x", b (hash [i]));
    j += 2;
  }
  rewinddir (cache_dir);
  int deleted = 0;
  struct dirent * dep = readdir (cache_dir);
  while (dep != NULL) {
    /* printf ("comparing %s to %s\n", dep->d_name, hash_name); */
    if (strcmp (dep->d_name, hash_name) == 0) {  /* match! */
      char fname [MAX_FNAME_LEN + 1];
      snprintf (fname, MAX_FNAME_LEN, "%s/%s", cache_dir_name, hash_name);
      struct stat file_info;
      if (stat (fname, &file_info) != 0) {
        perror ("stat of file to remove");
        printf ("unable to stat, should read header to determine length\n");
        file_info.st_size = 0;
        exit (1);  /* why unable to stat? */
      }
      printf ("unlinking %s\n", fname);
      if (unlink (fname) != 0) {
        perror ("unlink");
        printf ("unable to delete existing file %s\n", fname);
        return -1;
      }
      deleted++;
      used_storage -= strlen (fname) + file_info.st_size;
      if (used_storage < 0) {
        printf ("used_storage is %lld after subtracting %d and %ld\n",
                used_storage, strlen (fname), file_info.st_size);
        printf ("resetting used_storage to zero\n");
        used_storage = 0;
      }
    }
/* on to the next file */
    dep = readdir (cache_dir);
  }
  return deleted;
}

#ifdef PRINT_CACHE_CONTENTS

static void get_print_messages (struct allnet_addr * addr)
{
#define MAX_MSG		1000
  char * messages [MAX_MSG];
  int msizes [MAX_MSG];
  int priorities [MAX_MSG];
  int nm = get_messages (addr, 1, 1, messages, msizes, priorities, MAX_MSG);
  printf ("%d cached messages\n", nm);
  if (nm > MAX_MSG) nm = MAX_MSG;
#undef MAX_MSG
  int i;
  for (i = 0; i < nm; i++) {
    struct allnet_max_header h;
    struct allnet_header * hp = (struct allnet_header *) (&h);
    char * owner_id;
    int hsize = parse_header (messages [i], msizes [i], hp, &owner_id);
    if (hsize < 0) {
      printf ("unable to parse header of message %d (%d bytes)\n",
              i, msizes [i]);
    } else {
      printf ("message %d has %d bytes (%d in header), %d addresses in route",
              i, msizes [i], hsize, b (hp->allnet_num_addrs));
      printf (" (%d used)\n", b (hp->allnet_addrs_used));
      char * rp = messages [i] + ALLNET_HEADER_FIXED_SIZE;
      struct allnet_route route;
      int n = parse_route (rp, hsize - ALLNET_HEADER_FIXED_SIZE,
                           &route, b (hp->allnet_num_addrs), 
                           b (hp->allnet_addrs_used));
      print_route (&route, 1);
      free (route.addrs);
      free (messages [i]);
    }
  }
}

main (int argc, char ** argv)
{
  /* init_persist requires a maximum storage.  Since we won't be adding
   * anything to the cache, go ahead and give a large size for the cache */
  init_persist (1LL << 62, "allnet-cache");
  struct allnet_addr addr;
  addr.discrim = ALLNET_NO_ADDR_ID;
  get_print_messages (&addr);
}
#endif /* PRINT_CACHE_CONTENTS */

