/* persist.h: save a limited amount of data persistently, with priorities */

#ifndef ALLNET_PERSIST_H
#define ALLNET_PERSIST_H

#include "types.h"
#include "messages.h"

/* max_storage is in bytes, so is specified as a 64-bit number */
extern void init_persist (long long max_storage, char * id);

/* save a message, with the given priority.  A priority of zero doesn't
 * mean the message is deleted immediately, only, that the message has
 * the lowest possible priority.
 *
 * the header is used to determine the size of the message.  The
 * destination is assumed to be the last address in the route of the header.
 *
 * The priority may be set to decrease every second or every 24 hours.
 * if delta_priority_per_second is nonzero, the priority is
 * (logically) decremented every second.
 */
extern void save_message (char * msg,
                          struct allnet_header * header, /* parsed from msg */
                          int priority,
                          double delta_priority_per_second);

/* messages matching the first address in the route are returned in
 * order of priority.  Where priority is the same, older messages are
 * returned before newer messages.
 *
 * messages, message_sizes, and priorities should be array of at least
 * max_messages (char *) or ints
 * each entry of messages is set to point to malloc'd storage
 * containing the body of the message, should be free'd.
 *
 * the route is also used to build a valid message that can be sent.
 *
 * returns the number of messages, perhaps greater than max_messages
 * (but only the first max_messages are filled in)
 * may return 0 if there are no messages for this destination
 * will return -1 in case of errors
 */
extern int get_messages (struct allnet_addr * route, int num_addrs, int used,
                         char ** messages, int * message_sizes,
                         int * priorities, int max_messages);

/* delete any messages such that this owner ID hashes to their hash.
 * return the number of messages deleted, which may be 0, or -1 for errors.
 */
extern int delete_messages (char * owner_id, int osize);

#endif /* ALLNET_PERSIST_H */

