/* synchronized queue */
/* this queue synchronizes all accesses (sort of like a Hoare monitor)
 * so callers can perform calls in any order, and the messages are
 * correctly added to or removed from the queue.
 *
 * the queue is organized by priority, so elements with higher priority
 * are returned before elements with lower priority.  Objects with the
 * same priority are returned in the order of insertion.
 *
 * Objects are kept with their size as well as the time of insertion.
 * Objects may be removed from the queue either explicitly, or from
 * time to time by calling cleanup, which removes elements older than
 * a certain time and taking up too much space.
 *
 * it is assumed that elements are malloc'd, and any elements removed
 * from the queue (and not returned to the caller) will be free'd
 */

#include <sys/time.h>

#include "types.h"

struct syn_queue;

/* if optional_max_size is not zero, it is the maximum size, in bytes,
 * of all the elements in the queue.  Adding any element beyond the maximum
 * size results in the automatic removal of the lowest priority element(s),
 * until the maximum size is honored.  Objects larger than the maximum size
 * cannot be added to the queue.
 *
 * new_syn_queue returns NULL in case of error.
 */
extern struct syn_queue * new_syn_queue (int optional_max_size);
extern void free_syn_queue (struct syn_queue * q);

/* returns 1 if the queue is empty, 0 otherwise */
extern int queue_empty (struct syn_queue * q);

/* returns the number of bytes in elements stored in the queue */
extern int queue_size (struct syn_queue * q);

/* returns the number of elements stored in the queue */
extern int queue_elements (struct syn_queue * q);

/* returns 0 if the queue is empty, or the priority otherwise (note that
 * the highest priority could be zero).
 * If it does not return zero, and for each if it is not NULL, also fills in
 * the element value, size, and priority
 */
extern int queue_priority (struct syn_queue * q,
                           void ** value, int * size, int * priority);

/* sets *time the earliest time of any element stored in the queue, unless
 * the queue is empty or time is NULL.  If the queue is not empty,
 * and for each if it is not NULL, also fills in
 * the element value, size, and priority
 */
extern void queue_earliest (struct syn_queue * q, struct timeval * time,
                           void ** value, int * size, int * priority);

/* sets *time the latest time of any element stored in the queue, unless
 * the queue is empty or time is NULL.  If the queue is not empty,
 * and for each if it is not NULL, also fills in
 * the element value, size, and priority
 */
extern void queue_latest (struct syn_queue * q, struct timeval * time,
                          void ** value, int * size, int * priority);

/* size should be > 0, priority >= 0 */
/* returns the size in case of success,
 * -1 in case the element could not be added
 */
extern int queue_add (struct syn_queue * q,
                      void * element, int size, int priority);

/* returns (and removes) the first (highest priority) element in the queue,
 * or NULL if the queue is empty */
extern void * queue_get (struct syn_queue * q, int * size, int * priority);

/* same as queue_get, but if there is nothing in the queue, waits until
 * something is available, so never returns NULL. */
extern void * queue_wait (struct syn_queue * q, int * size, int * priority);

/* same as queue_get, but does not remove. */
extern void * queue_first (struct syn_queue * q, int * size, int * priority);

/* same as queue_first, but returns a malloc'd array of the first n values.
 * returns a value 0 <= i <= n giving the number of elements returned.
 * note the elements themselves are not copied, and are still in the queue.
/* use n = 1 for one element, 2 for the two elements, etc. */
/* the sizes and priorities arrays (unless sizes/priorities is/are NULL)
 * are also malloc'd, should be freed when no longer needed.
 */
extern int queue_n (struct syn_queue * q, int n,
                    void *** elements, int ** sizes, int ** priorities);

/* similar to queue_n, but (a) limits the returned elements to at most
 * size_limit total bytes, and (b) deletes the elements.  Also (c)
 * returns the number of elements and the actual size of the highest priority
 * elements that fit in the requested size. 
 */
extern void queue_n_size (struct syn_queue * q, int size_limit,
                          int * num_elements, int * actual_size,
                          void *** elements, int ** sizes,
                          int ** priorities);

/* applies the given function to the given argument, and every element
 * of the queue */
extern void queue_map (struct syn_queue * q, void * arg,
                       void (* fun) (void * v1, void * v2));

/* remove all elements for which the given function returns 1 when
 * applied to the given argument and a queue element.
 * Return the number of elements removed.  Removed element are freed */
extern int queue_remove (struct syn_queue * q, void * arg,
                         int (* match) (void * v1, void * v2));

