/* program to do the wireless sending and receiving for allnet */
/* this program runs in privileged mode so it can turn the wireless interface
 * on and off in ad-hoc mode.  It starts allnetd in non-privileged mode */
/* sending and receiving are on a 60s schedule which includes:
 * .3s (300ms) receiving, then 59.7s of a send loop:
 * .00125s (1.25ms) sending every .25s  (or 1.45ms every .29s)
 * sending the top priority messages that can be sent within the 1.25ms
 * may send longer if sending own data, but in any case not more than .125s
 * (half the cycle)
 *
 * on the test system as of 2012/09/12, it takes about 100ms to turn the
 * wireless interface on and off.  This time is included in the cycle
 * time, i.e. the cycle times above remain the same, so the duty cycle
 * is probably higher than 1%.   Fix later.
 *
 * after one 60s cycle, assume that the data is either delivered or
 * undeliverable, and request the next batch of data.
 *
 * do one .00125s sending somewhere at random during the .3s receiving, just
 * in case the peer's receive cycle is synchronized with ours.
 *
 * in general, the interface is set in the previous mode (whatever that was,
 * including off) while we are neither sending nor receiving.
 * 
 * the estimate for how long it takes to turn the wireless on and off is
 * included in the amount of data that can be sent in the 1.25ms send.
 *
 * if less than 1.25ms of data needs to be sent, can listen for the
 * remainder of the period
 */
/* The protocol for communicating with allnetd is simple:
 * give allnetd MAGIC_MESSAGE followed by the message for any received messages
 * give allnetd MAGIC_SENT followed by 64-bit byte count we may be able to send
 *  to signal we are ready for the next batch of data 
 * get from allnetd MAGIC_DATA followed by messages ending in a message
 *  with an allnet header indicating zero length (should be sane in case
 *  the message sizes exceed what was specified in MAGIC_SENT)
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <ifaddrs.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <net/ethernet.h>
#include <netpacket/packet.h>

#include "types.h"
#include "messages.h"
#include "wtxrxd.h"

static pid_t child_pid = 0;

struct interface_info {
  char * name;
  struct sockaddr_ll if_address; /* the address of the interface */
  struct sockaddr  * if_sap;     /* always points to if_address */
  struct sockaddr_ll bc_address; /* the broadcast address of the interface */
  struct sockaddr  * bc_sap;     /* always points to bc_address */
  int socket;
};
/* have to be global so htey is accessible to the signal handler */
static struct interface_info * interfaces = NULL;
static int num_wireless = 0;

static void print_buf (char * buffer, int bsize, int eol)
{
  int i;
  for (i = 0; i < bsize; i++) {
    printf ("%02x ", b (buffer [i]));
  }
  if ((eol) || (bsize > 30))
    printf ("\n");
}

static struct timeval add_us (struct timeval * t, unsigned int us)
{
#define US_PER_S	1000000
  struct timeval result;
  unsigned int result_us = t->tv_usec + us;
  result.tv_sec = t->tv_sec + (result_us / US_PER_S);
  result.tv_usec = (result_us % US_PER_S);
}

static unsigned long long delta_us (struct timeval * t1, struct timeval * t2)
{
  if ((t1->tv_sec < t2->tv_sec) ||
      ((t1->tv_sec == t2->tv_sec) &&
       (t1->tv_usec < t2->tv_usec)))  /* t1 before t2, return 0 */
    return 0LL;
  unsigned long long result = t1->tv_usec - t2->tv_usec;
  result += (t1->tv_sec - t2->tv_sec) * US_PER_S;
  return result;
}

static struct timeval delta_t (struct timeval * t1, struct timeval * t2)
{
  unsigned long long delta = delta_us (t1, t2);
  struct timeval result;
  result.tv_sec = delta / US_PER_S;
  result.tv_usec = delta % US_PER_S;
  return result;
}
#undef US_PER_S

static struct timeval random_middle (struct timeval * end)
{
  struct timeval now, middle;
  gettimeofday (&now, NULL);
  unsigned long long delta = delta_us (end, &now);
  if (delta == 0LL)
    return now;
  else
    return add_us (&now, ((unsigned int) (delta / 2)));
}

static int wireless_down ();

void wtxrxd_signal_handler (int signal)
{
  printf ("wtxrxd: got signal %d\n", signal);
  if (child_pid != 0) {
    if (kill (child_pid, signal) != 0) {
      perror ("kill in signal_handler");
      printf ("wtxrxd: unable to kill child process %d with signal %d\n",
              child_pid, signal);
    } else {
      printf ("wtxrxd: sent signal %d to child process %d\n",
              signal, child_pid);
    }
    unlink (ALLNETD_PID);
  }
  /* wireless_down is slow, but go ahead and do it when we shut down */
  /* wireless_down (); */
  exit (1);  /* terminate the process */
}

static void install_signal_handler ()
{
  /* catch control-C from the terminal */
  struct sigaction siga;
  siga.sa_handler = wtxrxd_signal_handler;
  sigemptyset (&(siga.sa_mask));
  siga.sa_flags = 0;
  if (sigaction (SIGINT, &siga, NULL) != 0) {
    perror ("sigaction/SIGINT");
    printf ("wtxrxd: unable to install signal handler for signal %d\n", SIGINT);
  }
  /* catch the kill signal from the kill command */
  if (sigaction (SIGTERM, &siga, NULL) != 0) {
    perror ("sigaction/SIGTERM");
    printf ("wtxrxd: unable to install signal handler for signal %d\n", SIGTERM);
  }
}

/* makes ALLNETD_PIPE_READ/WRITE into the right fds for the given pipes, */
/* closes the other two ends of the pipes */
static void make_my_pipes (int * pipes_to_me, int * pipes_from_me)
{
  if (dup2 (pipes_to_me [0], ALLNETD_PIPE_READ) != ALLNETD_PIPE_READ)
    perror ("dup2/1");
  if (dup2 (pipes_from_me [1], ALLNETD_PIPE_WRITE) != ALLNETD_PIPE_WRITE)
    perror ("dup2/2");
/* close write end of pipe to me */
  if (pipes_to_me [1] != ALLNETD_PIPE_WRITE)
    close (pipes_to_me [1]);
/* close read end of pipe from me */
  if (pipes_from_me [0] != ALLNETD_PIPE_READ)
    close (pipes_from_me [0]);
}

/* return 1 if successful, 0 otherwise */
static int my_execute (char * basic_command, char * interface, char * fail)
{
  int size = strlen (basic_command) + strlen (interface) + 1;
  char * command = malloc (size);
  if (command == NULL) {
    printf ("wtxrxd: unable to allocate %d bytes for command:\n", size);
    printf (basic_command, interface);
    return 0;
  }
  snprintf (command, size, basic_command, interface);
  int sys_result = system (command);
  /* printf ("wtxrxd: result of calling '%s' was %d\n", command, sys_result); */
  if (sys_result != 0) {
    if ((sys_result != -1) && (WIFEXITED (sys_result)))
      printf ("wtxrxd: program exit status for %s was %d\n",
              command, WEXITSTATUS (sys_result));
    if (fail != NULL)
      printf ("wtxrxd: call to '%s' failed, %s\n", command, fail);
    else
      printf ("wtxrxd: call to '%s' failed\n", command);
    return 0;
  }
  return 1;
}

static int one_wireless_up (char * interface)
{
/* need to execute the commands:
      sudo iw dev $if set type ibss
      sudo ifconfig $if up
      sudo iw dev $if ibss join allnet 2412
 */
  if (! my_execute ("iw dev %s set type ibss", interface,
                    "probably a wired or configured interface"))
    return 0;
  if (! my_execute ("ifconfig %s up", interface, NULL))
    return 0;
  if (! my_execute ("iw dev %s ibss join allnet 2412", interface, NULL))
    return 0;
  return 1;
}

static int wireless_up ()
{
  int i;
  for (i = 0; i < num_wireless; i++) {
    if (! one_wireless_up (interfaces [i].name))
      return 0;
  }
  return 1;
}

static int one_wireless_down (char * interface)
{
/* doesn't seem to be necessary or helpful
  if (! my_execute ("iw dev %s set type managed", interface, NULL))
    return 0;
*/
  if (! my_execute ("ifconfig %s down", interface, NULL))
    return 0;
  return 1;
}


static int wireless_down ()
{
/* need to execute the commands:
      sudo iw dev $if set type managed
      sudo ifconfig $if down
 */
  int i;
  for (i = 0; i < num_wireless; i++) {
    if (! one_wireless_down (interfaces [i].name))
      return 0;
  }
  return 1;
}

/* return 1 if the interface with the given name already exists, else 0 */
static int find_interface (char * name)
{
  int i;
  for (i = 0; i < num_wireless; i++)
    if (strcmp (name, interfaces [i].name) == 0)
      return 1;
  return 0;
}

static void add_to_list (struct ifaddrs * interface)
{
  if (interface->ifa_addr->sa_family != AF_PACKET) {  /* error */
    printf ("wtxrxd: illegal family %d\n", interface->ifa_addr->sa_family);
    exit (1);
  }

  int slen = strlen (interface->ifa_name) + 1;
  char * new_string = malloc (slen);
  if (new_string == NULL) {
    printf ("wtxrxd error: unable to allocate string for interface %s\n",
            interface->ifa_name);
    exit (1);  /* an initialization error, quit early */
  }
  strncpy (new_string, interface->ifa_name, slen);

  struct interface_info * new_interfaces = interfaces;
  int new_size = (num_wireless + 1) * sizeof (struct interface_info);
  if ((num_wireless == 0) || (interfaces == NULL))
    new_interfaces = malloc (new_size);
  else
    new_interfaces = realloc (new_interfaces, new_size);
  if (new_interfaces == NULL) {
    printf ("wtxrxd error: unable to allocate %d-th interface\n",
            num_wireless + 1);
    exit (1);  /* an initialization error, quit early */
  }
  interfaces = new_interfaces;

  struct interface_info * new_info = interfaces + num_wireless;
  new_info->name = new_string;
  new_info->if_address = *((struct sockaddr_ll *) (interface->ifa_addr));
  new_info->if_sap = (struct sockaddr *) (&(new_info->if_address));
  new_info->bc_sap = (struct sockaddr *) (&(new_info->bc_address));
  new_info->socket = socket (AF_PACKET, SOCK_DGRAM, ALLNET_WIFI_PROTOCOL);

  struct sockaddr_ll * if_sll = (struct sockaddr_ll *) (new_info->if_sap);
  struct sockaddr_ll * bc_sll = (struct sockaddr_ll *) (new_info->bc_sap);

  *bc_sll = *if_sll;
  if_sll->sll_family = AF_PACKET;
  if_sll->sll_protocol = 0; /* ALLNET_WIFI_PROTOCOL specified to socket(2) */
  if_sll->sll_hatype = 0;   /* not used */
  if_sll->sll_pkttype = 0;  /* not used */
  if_sll->sll_halen = 6;
  if_sll->sll_addr [0] = if_sll->sll_addr [1] = if_sll->sll_addr [2] =
    if_sll->sll_addr [3] = if_sll->sll_addr [4] = if_sll->sll_addr [5] = 0xff;

  num_wireless = num_wireless + 1;
}

/* send to allnetd a MAGIC_INIT followed by a single byte name length,
 * then the name of the interface to use */
static void send_init_to_allnetd (char * name)
{
  int slen = strlen (name);
  if (slen > 255) {
    printf ("interface %s: name too long (%d)\n", name, slen);
    exit (1);
  }
  printf ("wtxrxd: sending to allnetd interface name %s (%d bytes)\n",
          name, slen);
  char buf [MAGIC_SIZE + 257];
  memcpy (buf, MAGIC_INIT, MAGIC_SIZE);
  bs1 (buf + MAGIC_SIZE, slen);
  if (slen > 0)
    memcpy (buf + MAGIC_SIZE + 1, name, slen);
  int wsize = MAGIC_SIZE + 1 + slen;
  int w = write (ALLNETD_PIPE_WRITE, buf, wsize);
  if (w < wsize) {
    perror ("pipe write/unable to initialize allnetd\n");
    printf ("tried to write %d + 1 + %d = %d, wrote %d\n",
            MAGIC_SIZE, slen, wsize, w);
    exit (1);
  }
}

static int init_wireless ()
{
  static int first_interface = 1;
  num_wireless = 0;
  char * result = NULL;
  struct ifaddrs * ifa;
  if (getifaddrs (&ifa) != 0) {
    perror ("getifaddrs");
    exit (1);
  }
  struct ifaddrs * ifa_loop = ifa;
  while (ifa_loop != NULL) {
    if ((ifa_loop->ifa_addr->sa_family == AF_PACKET) &&
        (! find_interface (ifa_loop->ifa_name))) {
      printf ("wtxrxd: opening interface %s\n", ifa_loop->ifa_name);
      struct timeval start;
      gettimeofday (&start, NULL);
      if (one_wireless_up (ifa_loop->ifa_name)) {
        struct timeval midtime;
        gettimeofday (&midtime, NULL);
        long long mtime = delta_us (&midtime, &start);
        one_wireless_down (ifa_loop->ifa_name);
        struct timeval finish;
        gettimeofday (&finish, NULL);
        long long time = delta_us (&finish, &start);
        printf ("wtxrxd: %s is wireless, %lld.%03lld ms to turn on+off\n",
                ifa_loop->ifa_name, time / 1000LL, time % 1000LL);
        printf ("  (%lld.%03lld ms to turn on)\n",
                mtime / 1000LL, mtime % 1000LL);
        add_to_list (ifa_loop);
        if (first_interface) {
          send_init_to_allnetd (ifa_loop->ifa_name);
          first_interface = 0;
        }
      }
    }
    ifa_loop = ifa_loop->ifa_next;
  }
  freeifaddrs (ifa);
  int i;
  if (first_interface == 1)
    send_init_to_allnetd ("");
  printf ("wtxrxd: %d wireless, ", num_wireless);
  for (i = 0; i < num_wireless; i++)
    printf ("%s", interfaces [i].name);
  printf ("\n");
  return;
}

static void clear_pipe (int pipe)
{
  time_t start = time (NULL);
  struct timeval zero;
  fd_set read_fds;
  while (1) {
    zero.tv_sec = 0;
    zero.tv_usec = 0;
    FD_ZERO (&read_fds);
    FD_SET (pipe, &read_fds);
    if (select (pipe + 1, &read_fds, NULL, NULL, &zero) <= 0)
      return;    /* nothing to read, we are done */
    char buffer [1000];
    if (read (pipe, buffer, 1000) <= 0)
      return;    /* no data, or error, we are done */
    /* discard (by doing nothing) any data we received, and repeat */
    if (time (NULL) - start > 2) {
      printf ("wtxrxd error: more than 2 seconds to clear pipe\n");
      return;
    }
  }
}

static void sleep_until (struct timeval * stop)
{
  struct timeval now;
  gettimeofday (&now, NULL);
  struct timeval sleep = delta_t (stop, &now);
  if ((sleep.tv_usec == 0) && (sleep.tv_sec == 0))
    return;
  select (0, NULL, NULL, NULL, &sleep);
}

static int read_until (int fd, char * buffer, int bsize, struct timeval * stop)
{
  if (bsize <= 0)
    return 0;
  fd_set read_fds;
  FD_ZERO (&read_fds);
  FD_SET (fd, &read_fds);
  struct timeval now;
  gettimeofday (&now, NULL);
  struct timeval sleep = delta_t (stop, &now);
  if (select (fd + 1, &read_fds, NULL, NULL, &sleep) == 0)
    return 0;   /* timeout */
  int size = read (fd, buffer, bsize);
  if (size < 0) {
    perror ("wtxrxd read data");
    return 0;
  }
  return size;
}

/* return 1 for successful read, 0 otherwise */
/* take no more than 10ms to finish reading */
static int read_data_info (char * data_buffer, int send_per_cycle,
                           long long * send_messages, long long * send_size)
{
  printf ("wtxrxd: rd_info called, %d max\n", send_per_cycle);
  char send_info [MAGIC_SIZE + 16];
  int offset = 0;
  int found_magic = 0;
  struct timeval now;
  gettimeofday (&now, NULL);
  struct timeval finish = add_us (&now, 10000 + 1000000 /* 1s */ );
  do {
    int hsize = read_until (ALLNETD_PIPE_READ, send_info + offset,
                            sizeof (send_info) - offset, &finish);
    if (hsize == 0) {
      if (offset != 0)
        printf ("wtxrxd: rd_info read %d header bytes, but now returning\n",
                offset);
      else
        printf ("wtxrxd: rd_info read nothing in 10ms\n");
      return 0;
    }
    offset += hsize;
    found_magic = (memcmp (send_info, MAGIC_DATA, MAGIC_SIZE) == 0);
    while ((offset >= MAGIC_SIZE) && (! found_magic)) {
      memmove (send_info, send_info + 1, offset - 1);
      offset--;
      found_magic = (memcmp (send_info, MAGIC_DATA, MAGIC_SIZE) == 0);
    }
  } while (offset != sizeof (send_info));
  *send_messages = b8 (send_info + MAGIC_SIZE);
  *send_size = b8 (send_info + MAGIC_SIZE + 8);
  if ((*send_size > send_per_cycle) ||
      (*send_messages / ALLNET_HEADER_FIXED_SIZE > *send_size)) {
    printf ("wtxrxd: rd_info max %d, but size %lld, %lld messages (>= %d)\n",
            send_per_cycle, *send_size, *send_messages,
            ALLNET_HEADER_FIXED_SIZE);
    return 0;
  }
  offset = 0;
  do {
    int msize = read_until (ALLNETD_PIPE_READ, data_buffer + offset,
                            *send_size - offset, &finish);
    if (msize == 0) {
      printf ("wtxrxd: rd_info read 0 bytes (%d/%lld), returning %lld - %lld\n",
              offset, (*send_size) - offset, *send_messages, *send_size); 
      return offset;
    }
    offset += msize;
  } while (offset < *send_size);
  printf ("wtxrxd: rd_info returning %lld messages, total size %lld\n",
          *send_messages, *send_size); 
  return 1;
}

/* send all the packets once on all the interfaces */
static void send_once (char * send_buffer,
                       long long num_messages, long long total_size)
{
  long long offset = 0;
  int message;
  for (message = 0; message < num_messages; message++) {
    struct allnet_max_header header;
    struct allnet_header * hp = (struct allnet_header *) (&header);
    if (parse_header (send_buffer + offset, total_size - offset,
                      hp, NULL) < ALLNET_HEADER_FIXED_SIZE) {
      printf ("wtxrxd: send_once unable to parse header for %d of %lld\n",
              message, num_messages);
      return;  /* not much else we can do */
    }
    long long size = hp->total_length;
    if (size + offset > total_size) {
      printf ("wtxrxd: send_once [%d]: offset %lld + size %lld > buf %lld\n",
              message, offset, size, total_size);
      return;
    }
    int interface;
    for (interface = 0; interface < num_wireless; interface++) {
      struct sockaddr_ll      unicast;  /* declared here in case we need it */
      struct sockaddr       * use_addr = interfaces [interface].bc_sap;
      if (hp->allnet_addrs_used < hp->allnet_num_addrs) {
        struct allnet_addr * next = hp->route + b (hp->allnet_addrs_used);
        if ((next->discrim == ALLNET_UNICAST_ADDR_ID) &&
            ((b (next->uc_addr.num_bits)) >= 48)) {
          unicast = interfaces [interface].bc_address;
          unicast.sll_addr [0] = b1 (next->pk_addr.pubkey + 0);
          unicast.sll_addr [1] = b1 (next->pk_addr.pubkey + 1);
          unicast.sll_addr [2] = b1 (next->pk_addr.pubkey + 2);
          unicast.sll_addr [3] = b1 (next->pk_addr.pubkey + 3);
          unicast.sll_addr [4] = b1 (next->pk_addr.pubkey + 4);
          unicast.sll_addr [5] = b1 (next->pk_addr.pubkey + 5);
          use_addr = (struct sockaddr *) (&unicast);
        }
      }
      int adrlen = sizeof (struct sockaddr_ll);
      if (sendto (interfaces [interface].socket, send_buffer + offset, size,
                  MSG_DONTWAIT, use_addr, adrlen) < size)
        perror ("sento/send_once");
      offset += size;
    }
  }
  if (offset != total_size) {
    printf ("wtxrxd: sent %lld packets with size %lld, buffer size %lld\n",
            num_messages, offset, total_size);
  }
}

static void send_until (char * send_buffer, long long num_messages,
                        long long total_size, struct timeval * stop)
{
/* send .00125s (1.25ms) every .25s -- the amount of data should take
 * care of the sending 1.25ms, so here we only keep track of how often
 * to send and when to stop sending. */
  while (1) {
    struct timeval cycle_start;
    gettimeofday (&cycle_start, NULL);
    if (delta_us (stop, &cycle_start) <= 0)
      break;
    struct timeval cycle_end = add_us (&cycle_start, 250 * 1000);
    if (delta_us (stop, &cycle_end) <= 0)
      cycle_end = *stop;
    wireless_up ();
    send_once (send_buffer, num_messages, total_size);
    wireless_down ();
    sleep_until (&cycle_end);
  }
}

#define MAX_RECEIVE_BUFFER	65507	/* max UDP size */
static void do_receive_until (char * buf, struct timeval * stop)
{
  while (1) {
    struct timeval now;
    gettimeofday (&now, NULL);
    struct timeval sleep = delta_t (stop, &now);
    if ((sleep.tv_usec == 0) && (sleep.tv_sec == 0))  /* finished */
      return;
    fd_set read_fds;
    FD_ZERO (&read_fds);
    int max = 0;
    int i;
    for (i = 0; i < num_wireless; i++) {
      if (max <= interfaces [i].socket)
        max = interfaces [i].socket + 1;
      FD_SET (interfaces [i].socket, &read_fds);
    }
    int s = select (max, &read_fds, NULL, NULL, &sleep);
    if (s < 0)
      perror ("select/do_receive_until");
    if (s <= 0)    /* error or timeout, we are done */
      return;

    for (i = 0; i < num_wireless; i++) {
      if (! FD_ISSET (interfaces [i].socket, &read_fds))
        continue;    /* check the next interface */
/*  we don't really use the sender address
 *    struct sockaddr_storage sas;
 *    struct sockaddr * sap = (struct sockaddr *) (&sas);
 *    int adrlen = sizeof (sas);
 *    int r = recvfrom (interfaces [i].socket, buf, MAX_RECEIVE_BUFFER,
 *                      MSG_DONTWAIT, sap, &adrlen);
 */
      int r = recvfrom (interfaces [i].socket, buf, MAX_RECEIVE_BUFFER,
                        MSG_DONTWAIT, NULL, NULL);
      if (r < 0)
        perror ("recvfrom/do_receive_until");
      if (r <= 0)
        continue;  /* did not receive anything, try the next interface */

      int w = write (ALLNETD_PIPE_WRITE, MAGIC_MESSAGE, MAGIC_SIZE);
      if (w != MAGIC_SIZE) {
        perror ("write received message header to allnetd");
        printf ("wtxrxd error: unable to send header of size %d, sent %d\n",
                MAGIC_SIZE, w);
      } else {
        w = write (ALLNETD_PIPE_WRITE, buf, r);
        if (w != r) {
          perror ("write received message to allnetd");
          printf ("wtxrxd error: unable to send message of size %d, sent %d\n",
                  r, w);
        }
      }
    }
  }
}

/* receive until the given time.  At a random point during the middle
 * part of the interval and unless send_buffer is NULL, call send_once.
 */
static void receive_until (char * buf, struct timeval * stop,
                           char * send_buffer, long long num_messages,
                           long long total_size)
{
  /* receive until the stop time, with a send somewhere in between */
  wireless_up ();
  struct timeval send_time;
  if (send_buffer != NULL) {
    struct timeval send_time;
    send_time = random_middle (stop);
    do_receive_until (buf, &send_time);
    send_once (send_buffer, num_messages, total_size);
  }
  do_receive_until (buf, stop);
  wireless_down ();
}

/* sending and receiving are on a 60s schedule which includes:
 * .3s (300ms) receiving, then 59.7s of a send loop:
 * .00125s (1.25ms) sending every .25s  (or 1.45ms every .29s)
 * sending the top priority messages that can be sent within the 1.25ms
 * may send longer if sending own data, but in any case not more than .125s
 * (half the cycle)
 *
 * after one 60s cycle, assume that the data is either delivered or
 * undeliverable, and request the next batch of data.
 *
 * the data is requested from allnetd at the beginning of the cycle
 */
static int main_loop (int send_per_cycle)
{
  int bsize = send_per_cycle + MAX_RECEIVE_BUFFER;
  char * buffer_space = malloc (bsize);
  if (buffer_space == NULL) {
    printf ("wtxrxd: malloc (%d) failed, aborting\n", bsize);
    exit (1);
  }
  char * send_buffer; /* = buffer_space or NULL if no data to send */
  int send_messages;
  int send_size;
  char * receive_buffer = buffer_space + send_per_cycle;

  char request_from_allnetd [MAGIC_SIZE + 8];
  memcpy (request_from_allnetd, MAGIC_SENT, MAGIC_SIZE);
  bs8 (request_from_allnetd + MAGIC_SIZE, ((long long) send_per_cycle));

  /* wireless_up is slow -- just go ahead and turn it on for now */
  /* wireless_up (); */
  struct timeval cycle_start;
  while (1) {
    clear_pipe (ALLNETD_PIPE_READ);
    /* send to allnetd the request for more data */
    int b = write (ALLNETD_PIPE_WRITE, request_from_allnetd,
                   sizeof (request_from_allnetd));
    if (b != sizeof (request_from_allnetd)) {
      perror ("send data size to allnetd");
      printf ("wtxrxd error: unable to send number of bytes %d\n",
              send_per_cycle);
    }
    long long send_messages = 0LL;
    long long send_size = 0LL;
    send_buffer = buffer_space;
    if (! read_data_info (send_buffer, send_per_cycle,
                          &send_messages, &send_size))
      send_buffer = NULL;

    gettimeofday (&cycle_start, NULL);
    struct timeval receive_cycle_end = add_us (&cycle_start, 300000);
    receive_until (receive_buffer, &receive_cycle_end,
                   send_buffer, send_messages, send_size);
    struct timeval next_cycle_start = cycle_start;
    next_cycle_start.tv_sec += 60;
    if (send_buffer != NULL)
      send_until (send_buffer, send_messages, send_size, &next_cycle_start);
    else
      sleep_until (&next_cycle_start);
  }
}

int main (int argc, char ** argv)
{
  /* should be done first, before any use of stdout */
  if (setvbuf (stdout, NULL, _IONBF, 0) != 0) {
    perror ("setvbuf");
    printf ("wtxrxd: unable to set stdout to unbuffered mode\n");
  }
  if (geteuid () != 0) {
    printf ("wtxrxd must be executed as root, running as user %d, aborting\n",
            geteuid ());
    exit (1);
  }
  install_signal_handler ();
  /* see if allnetd is running */
  int running = 0;
  int pid_fd = open (ALLNETD_PID, O_RDWR);
  if (pid_fd >= 0) {  /* file exists */
    char pid_value [100];
    int bytes = read (pid_fd, pid_value, sizeof (pid_value) - 1);
    if (bytes > 0) {
      pid_value [bytes] = '\0';
      int matched = sscanf (pid_value, "%d", &child_pid);
      if ((matched == 1) && (child_pid > 0) && (kill (child_pid, 0) == 0))
        running = 1;
    }
    close (pid_fd);
  }
  if (! running) {
    int log_fd = open (ALLNETD_LOG, O_WRONLY | O_CREAT | O_APPEND, 0644);
    if (log_fd < 0) {
      perror ("open log\n");
      printf ("wtxrxd: unable to open log file %s, output to terminal\n",
              ALLNETD_LOG, log_fd);
    } else {
#define OUT_FD	1
      if (close (OUT_FD) != 0) {
        perror ("close");
      }
      if (dup2 (log_fd, OUT_FD) != OUT_FD) { 
     /* redirect standard output to the log */
        perror ("dup2");
      }
      close (log_fd);
    }
    int pipes_allnetd_to_wtxrxd [2];
    if (pipe (pipes_allnetd_to_wtxrxd) != 0) {
      perror ("pipe/1");
    }
    int pipes_wtxrxd_to_allnetd [2];
    if (pipe (pipes_wtxrxd_to_allnetd) != 0) {
      perror ("pipe/2");
    }
    printf ("wtxrxd: pipes allnetd to wtxrxd are: %d read, %d write\n",
            pipes_allnetd_to_wtxrxd [0], pipes_allnetd_to_wtxrxd [1]);
    printf ("wtxrxd: pipes wtxrxd to allnetd are: %d read, %d write\n",
            pipes_wtxrxd_to_allnetd [0], pipes_wtxrxd_to_allnetd [1]);
    child_pid = fork ();
    if (child_pid == 0) {  /* child -- change uid and exec allnetd */
      make_my_pipes (pipes_wtxrxd_to_allnetd, pipes_allnetd_to_wtxrxd); 
      if (setuid (65534) != 0) {      /* nobody */
        perror ("setuid");
        printf ("wtxrxd: unable to setuid to nobody, not starting allnetd\n");
        exit (1);
      }
      char * const argv [] = { ALLNETD_PROGRAM, NULL };
      execv (ALLNETD_PATH, argv);
      perror ("execv");
      printf ("wtxrxd: unable to execute %s\n", ALLNETD_PATH);
      exit (1);
    }
    pid_fd = open (ALLNETD_PID, O_RDWR | O_CREAT | O_TRUNC, 0644);
    char pid_value [100];
    snprintf (pid_value, sizeof (pid_value), "%d\n", child_pid);
    write (pid_fd, pid_value, strlen (pid_value));
    close (pid_fd);
    printf ("wtxrxd: created child process %s, pid %d\n",
            ALLNETD_PATH, child_pid);
    make_my_pipes (pipes_allnetd_to_wtxrxd, pipes_wtxrxd_to_allnetd); 
  } else {
    printf ("wtxrxd: allnetd is already running, not started\n");
  }
  init_wireless ();
#define DEBUG_SIZE	10000LL
  main_loop (DEBUG_SIZE);
}
